"""
This script lists package constants.
"""


NO_ACCESS_RIGHT = '[noAccessRights] It seems you do not have access to this object.' \
    ' Please double check your credentials.'

mime_extensions = {
  'Audio': 'audio/x-flac,audio/mpeg,video/mp4',
  'Csv': 'text/csv',
  'Frame': 'video/mp4,video/x-matroska,video/3gpp,video/x-msvideo,' \
    'video/x-m4v,video/quicktime,video/webm',
  'Image': 'image/jpeg,image/png,image/bmp,image/gif,image/webp,image/x-icon,'\
      'image/tiff,image/vnd.microsoft.icon,image/svg+xml,image/avif,image/apng',
  'Pdf': 'application/pdf',
  'Text': 'text/plain',
  'TimeSeries': 'text/csv',
}

mime_extensions_for_IV2 = {
  'AUDIO': mime_extensions['Audio'],
  'FRAME': mime_extensions['Frame'],
  'IMAGE': mime_extensions['Image'],
  'NA': '',
  'PDF': mime_extensions['Pdf'],
  'TEXT': mime_extensions['Text'],
  'TIME_SERIES': mime_extensions['Csv'],
  'URL': '',
  'VIDEO': mime_extensions['Csv'],
}

MUTATION_BATCH_SIZE = 100
THROTTLING_DELAY = 60/250



"""
Helpers for the asset mutations
"""
import csv
import os
from json import dumps
from uuid import uuid4
from typing import List, Union
import mimetypes

from ...constants import mime_extensions_for_IV2
from ...helpers import (convert_to_list_of_none, encode_base64, format_metadata, \
    get_data_type, is_none_or_empty, is_url)
from .queries import (GQL_APPEND_MANY_TO_DATASET,
                      GQL_APPEND_MANY_FRAMES_TO_DATASET)


def encode_object_if_not_url(content, input_type):
    """
    Return the object if it is a url, else it should be a path to a file.
    In that case, the file is returned as a base64 string
    """
    if is_url(content):
        return content
    if check_file_mime_type(content, input_type):
        return encode_base64(content)
    return None


def process_frame_json_content(json_content):
    """
    Function to process individual json_content of FRAME projects
    """
    if is_url(json_content):
        return json_content
    json_content_index = range(len(json_content))
    json_content_urls = [encode_object_if_not_url(content, 'IMAGE') for content in json_content]
    return dumps(dict(zip(json_content_index, json_content_urls)))


def get_file_mimetype(content_array: Union[List[str], None],
                      json_content_array: Union[List[str], None]) -> Union[str, None]:
    """
    Returns the mimetype of the first file of the content array
    """
    if json_content_array is not None:
        return None
    if content_array is None:
        return None
    if len(content_array) > 0:
        first_asset = content_array[0]
        if is_url(first_asset):
            return None
        if not os.path.exists(first_asset):
            return None
        return mimetypes.guess_type(first_asset)[0]
    return None


def process_json_content(input_type: str,
                         content_array: List[str],
                         json_content_array: Union[List[str], None]):
    """
    Process the array of json_contents
    """
    if json_content_array is None:
        return [''] * len(content_array)
    if input_type == 'FRAME':
        return list(map(process_frame_json_content, json_content_array))
    return [element if is_url(element) else dumps(element) for element in json_content_array]


def process_content(input_type: str,
                    content_array: Union[List[str], None],
                    json_content_array: Union[List[List[Union[dict, str]]], None]):
    """
    Process the array of contents
    """
    if input_type in ['IMAGE', 'PDF']:
        return [content if is_url(content) else (content
            if (json_content_array is not None and json_content_array[i] is not None)
            else (encode_base64(content) if check_file_mime_type(content, input_type) else None))
            for i, content in enumerate(content_array)]
    if input_type == 'FRAME' and json_content_array is None:
        content_array = [encode_object_if_not_url(content, input_type) for content in content_array]
    if input_type == 'TIME_SERIES':
        content_array = list(map(process_time_series, content_array))
    return content_array

def process_time_series(content: str) -> Union[str, None]:
    """
    Process the content for TIME_SERIES projects: if it is a file, read the content
    and also check if the content corresponds to the expected format, else return None
    """
    delimiter = ','
    if os.path.isfile(content):
        if check_file_mime_type(content, 'TIME_SERIES'):
            with open(content, 'r', encoding='utf8') as csvfile:
                reader = csv.reader(csvfile, delimiter=delimiter)
                return process_csv_content(reader, file_name=content, delimiter=delimiter)
        return None

    reader = csv.reader(content.split('\n'), delimiter=',')
    return process_csv_content(reader, delimiter=delimiter)

def process_csv_content(reader, file_name = None, delimiter=',') -> bool:
    """
    Process the content of csv for time_series and check if it corresponds to the expected format
    """
    first_row = True
    processed_lines = []
    for row in reader:
        if not (len(row)==2 and (first_row or (not first_row and is_float(row[0])))):
            print(f"""The content {file_name if file_name else row} does not correspond to the \
correct format: it should have only 2 columns, the first one being the timestamp \
(an integer or a float) and the second one a numeric value (an integer or a float, \
otherwise it will be considered as missing value). The first row should have the names \
of the 2 columns. The delimiter used should be ','.""")
            return None
        value = row[1] if (is_float(row[1]) or first_row) else ''
        processed_lines.append(delimiter.join([row[0], value]))
        first_row = False
    return '\n'.join(processed_lines)

def is_float(number: str) -> bool:
    """
    Check if a string can be converted to float
    """
    try:
        float(number)
        return True
    except ValueError:
        return False

def check_file_mime_type(content: str, input_type: str) -> bool:
    """
    Returns true if the mime type of the file corresponds to the allowed mime types of the project
    """
    if input_type not in ['IMAGE', 'FRAME', 'PDF', 'TIME_SERIES']:
        return True

    mime_type = get_data_type(content.lower())

    if not (mime_extensions_for_IV2[input_type] and mime_type):
        return False

    correct_mime_type = mime_type in mime_extensions_for_IV2[input_type]
    if not correct_mime_type:
        print(f'File mime type for {content} is {mime_type} and does not correspond' \
            'to the type of the project. '\
            f'File mime type should be one of {mime_extensions_for_IV2[input_type]}')
    return correct_mime_type

def add_video_parameters(json_metadata, should_use_native_video):
    """
    Add necessary video parameters to the metadata of the video
    """
    processing_parameters = json_metadata.get('processingParameters', {})
    video_parameters = [('shouldKeepNativeFrameRate', should_use_native_video), (
        'framesPlayedPerSecond', 30), ('shouldUseNativeVideo', should_use_native_video)]
    for (key, default_value) in video_parameters:
        processing_parameters[key] = processing_parameters.get(
            key, default_value)
    return {**json_metadata, 'processingParameters': processing_parameters}


def process_metadata(input_type: str, content_array: Union[List[str], None],
                     json_content_array: Union[List[List[Union[dict, str]]], None],
                     json_metadata_array: Union[List[dict], None]):
    """
    Process the metadata of each asset
    """
    json_metadata_array = [
        {}] * len(content_array) if json_metadata_array is None else json_metadata_array
    if input_type == 'FRAME':
        should_use_native_video = json_content_array is None
        json_metadata_array = [add_video_parameters(
            json_metadata, should_use_native_video) for json_metadata in json_metadata_array]
    return list(map(dumps, json_metadata_array))


def get_request_to_execute(
    input_type: str,
    json_metadata_array: Union[List[dict], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    mime_type: Union[str, None]
) -> str:
    """
    Selects the right query to run versus the data given
    """
    if json_content_array is not None:
        return GQL_APPEND_MANY_TO_DATASET, None
    if input_type != 'FRAME':
        if input_type == 'IMAGE' and mime_type == 'image/tiff':
            return GQL_APPEND_MANY_FRAMES_TO_DATASET, 'GEO_SATELLITE'
        return GQL_APPEND_MANY_TO_DATASET, None
    if json_metadata_array is None:
        return GQL_APPEND_MANY_TO_DATASET, None
    if (isinstance(json_metadata_array, list) and
        len(json_metadata_array) > 0 and
        not json_metadata_array[0].get(
            'processingParameters', {}).get('shouldUseNativeVideo', True)):
        return GQL_APPEND_MANY_FRAMES_TO_DATASET, 'VIDEO'
    return GQL_APPEND_MANY_TO_DATASET, None


# pylint: disable=too-many-arguments
def process_append_many_to_dataset_parameters(
        input_type: str,
        content_array: Union[List[str], None],
        external_id_array: Union[List[str], None],
        is_honeypot_array: Union[List[str], None],
        status_array: Union[List[str], None],
        json_content_array: Union[List[List[Union[dict, str]]], None],
        json_metadata_array: Union[List[dict], None]
):
    """
    Process arguments of the append_many_to_dataset method and return the data payload.
    """
    if content_array is None and json_content_array is None:
        raise ValueError(
            "Variables content_array and json_content_array cannot be both None.")
    if content_array is None:
        content_array = [''] * len(json_content_array)
    if external_id_array is None:
        external_id_array = [
            uuid4().hex for _ in range(len(content_array))]
    is_honeypot_array = [
        False] * len(content_array) if is_honeypot_array is None else is_honeypot_array
    status_array = ['TODO'] * \
        len(content_array) if not status_array else status_array
    formatted_json_metadata_array = process_metadata(
        input_type, content_array, json_content_array, json_metadata_array)
    mime_type = get_file_mimetype(content_array, json_content_array)
    content_array = process_content(
        input_type, content_array, json_content_array)
    formatted_json_content_array = process_json_content(
        input_type, content_array, json_content_array)

    request, upload_type = get_request_to_execute(
        input_type, json_metadata_array, json_content_array, mime_type)
    properties = {
        'content_array': content_array,
        'external_id_array': external_id_array,
        'is_honeypot_array': is_honeypot_array,
        'status_array': status_array,
        'json_content_array': formatted_json_content_array,
        'json_metadata_array': formatted_json_metadata_array,
    }

    return properties, upload_type, request

def process_update_properties_in_assets_parameters(properties) -> dict:
    """
    Process arguments of the update_properties_in_assets method
    and return the properties for the paginated loop
    """
    formatted_json_metadatas = None
    if properties['json_metadatas'] is None:
        formatted_json_metadatas = None
    else:
        if isinstance(properties['json_metadatas'], list):
            formatted_json_metadatas = list(
                map(format_metadata, properties['json_metadatas']))
        else:
            raise Exception('json_metadatas',
                            'Should be either a None or a list of None, string, list or dict')
    properties['json_metadatas'] = formatted_json_metadatas
    nb_assets_to_modify = len(properties['asset_ids'])
    properties = {k: convert_to_list_of_none(
        v, length=nb_assets_to_modify) for k, v in properties.items()}
    properties['should_reset_to_be_labeled_by_array'] = list(map(
        is_none_or_empty, properties['to_be_labeled_by_array']))
    return properties
    


"""User mutations."""

from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from .queries import (GQL_CREATE_USER,
                      GQL_RESET_PASSWORD,
                      GQL_UPDATE_PASSWORD, GQL_UPDATE_PROPERTIES_IN_USER)


class MutationsUser:
    """Set of User mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def create_user(self,
                    email: str = None,
                    password: str = None,
                    organization_role: str = None,
                    firstname: Optional[str] = None,
                    lastname: Optional[str] = None):
        """Add a user to your organization.

        Args:
            email: Email of the new user, used as user's unique identifier.
            password: On the first sign in, he will use this password and be able to change it.
            organization_role: One of "ADMIN", "USER".
            firstname: First name of the new user.
            lastname: Last name of the new user.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            'data': {'email': email,
                     'password': password,
                     'organizationRole': organization_role}
        }
        if firstname is not None:
            variables['data']['firstname'] = firstname
        if lastname is not None:
            variables['data']['lastname'] = lastname
        result = self.auth.client.execute(GQL_CREATE_USER, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_password(self, email: str, old_password: str, new_password_1: str,
                        new_password_2: str):
        """Allow to modify the password that you use to connect to Kili. \
        This resolver only works for on-premise installations without Auth0.

        Args:
            email : Email of the person whose password has to be updated.
            old_password : The old password
            new_password_1: The new password
            new_password_2: A confirmation field for the new password

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            'data': {'oldPassword': old_password,
                     'newPassword1': new_password_1,
                     'newPassword2': new_password_2},
            'where': {'email': email}
        }
        result = self.auth.client.execute(GQL_UPDATE_PASSWORD, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def reset_password(self, email: str):
        """Reset password.

        This resolver only works for on-premise installations without Auth0,
        if your organization allows Kili to send emails.

        Args:
            email: Email of the person whose password has to be reset.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {'where': {'email': email}}
        result = self.auth.client.execute(GQL_RESET_PASSWORD, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_properties_in_user(self,
                                  email: str,
                                  firstname: Optional[str] = None,
                                  lastname: Optional[str] = None,
                                  organization_id: Optional[str] = None,
                                  organization_role: Optional[str] = None,
                                  activated: Optional[bool] = None):
        """Update the properties of a user.

        Args:
            email: The email is the identifier of the user.
            firstname:Change the first name of the user.
            lastname: Change the last name of the user.
            organization_id: Change the organization the user is related to.
            organization_role: Change the role of the user.
                One of "ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER".
            activated: In case we want to deactivate a user, but keep it.


        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            'email': email,
        }
        if firstname is not None:
            variables['firstname'] = firstname
        if lastname is not None:
            variables['lastname'] = lastname
        if organization_id is not None:
            variables['organizationId'] = organization_id
        if organization_role is not None:
            variables['organizationRole'] = organization_role
        if activated is not None:
            variables['activated'] = activated
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_USER, variables)
        return format_result('data', result)



"""
Test notebooks with pytest
"""

import os
import subprocess

import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError

RECIPES_DIR = './recipes'
RECIPES_TESTED = [
    'create_project.ipynb',
    'export_labels.ipynb',
    'frame_dicom_data.ipynb',
    'import_assets.ipynb',
    'import_predictions.ipynb',
    'import_text_assets.ipynb',
    'inference_labels.ipynb',
    'medical_imaging.ipynb',
    'ocr_pre_annotations.ipynb',
    'pixel_level_masks.ipynb',
    'query_methods.ipynb',
    'webhooks.ipynb',
    'getting-started/getting_started-image_bounding_box_detection.ipynb',
    'getting-started/getting_started-image_classification.ipynb',
    'getting-started/getting_started-image_semantic_segmentation.ipynb',
    'getting-started/getting_started-named-entity-recognition.ipynb',
]

def process_notebook(notebook_filename):
    """
    Checks if an IPython notebook runs without error from start to finish.
    """
    with open(notebook_filename) as f:
        nb = nbformat.read(f, as_version=4)
    
    ep = ExecutePreprocessor(timeout=1000, kernel_name='python3')

    try:
        # Check that the notebook runs
        ep.preprocess(nb, {'metadata': {'path': ''}})
    except CellExecutionError:
        raise
         
    print(f"Successfully executed {notebook_filename}")
    return
    
def test_all_recipes():
    """
    Runs `process_notebook` on all notebooks in the git repository.
    """
    notebooks = [os.path.join(RECIPES_DIR, recipe) for recipe in RECIPES_TESTED]
    for notebook in notebooks:
        print("Testing", notebook)
        process_notebook(notebook)
        
    return

if __name__ == '__main__':
    test_all_recipes()
    


from datetime import datetime, timedelta
from functools import wraps

COUNT_SAMPLE_MAX = 26000


class burstthrottle(object):
    """
    Decorator that prevents a function from being called more that a certain amount of time
    To create a function that cannot be called more than 250 times in a minute:
        @burstthrottle(max_hits = 250, minutes = 1)
        def my_fun():
            pass
    """

    def __init__(self, max_hits, seconds=0, minutes=1, hours=0, error_message='TooManyCalls'):
        self.burst_window = timedelta(
            seconds=seconds, minutes=minutes, hours=hours)
        self.error_message = error_message
        self.hits = 0
        self.max_hits = max_hits
        self.timestamp = datetime.min

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            now = datetime.now()
            if now < self.timestamp + self.burst_window:
                if self.hits < self.max_hits:
                    self.hits += 1
                    return fn(*args, **kwargs)
                else:
                    self.timestamp = datetime.min
                    self.hits = 0
                    raise ValueError(self.error_message)

            else:
                self.timestamp = datetime.now()
                self.hits = 1
                return fn(*args, **kwargs)

        return wrapper


@burstthrottle(max_hits=250, minutes=1)
def mocked_query_method(skip, first, *_):
    """
    Simulates a query result by returning a list of ids
    """
    max_range = min(COUNT_SAMPLE_MAX, skip + first)
    res = [{"id": i} for i in range(skip, max_range)]
    return res


@burstthrottle(max_hits=250, minutes=1)
def mocked_count_method(*_):
    """
    Simulates a count query
    """
    return COUNT_SAMPLE_MAX



"""Notification queries."""

from typing import Generator, List, Optional, Union
import warnings

from typeguard import typechecked


from ...helpers import Compatible, format_result, fragment_builder
from .queries import gql_notifications, GQL_NOTIFICATIONS_COUNT
from ...types import Notification
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesNotification:
    """Set of Notification queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v2'])
    @typechecked
    def notifications(self,
                      fields: List[str] = ['createdAt', 'hasBeenSeen',
                                           'id', 'message', 'status', 'userID'],
                      first: int = 100,
                      has_been_seen: Optional[bool] = None,
                      notification_id: Optional[str] = None,
                      skip: int = 0,
                      user_id: Optional[str] = None,
                      disable_tqdm: bool = False,
                      as_generator: bool = False) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of notifications respecting a set of criteria.

        Args:
            fields: All the fields to request among the possible fields for the notifications
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#notification) for all possible fields.
            first: Number of notifications to query
            has_been_seen: If the notifications returned should have been seen.
            notification_id: If given, will return the notification which has this id
            skip: Number of notifications to skip (they are ordered by their date of creation,
                first to last).
            user_id: If given, returns the notifications of a specific user
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the notifications is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """

        count_args = {"has_been_seen": has_been_seen, "user_id": user_id}
        disable_tqdm = disable_tqdm or as_generator or notification_id is not None
        payload_query = {
            'where': {
                'id': notification_id,
                'user': {
                    'id': user_id,
                },
                'hasBeenSeen': has_been_seen,
            },
        }
        notifications_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_notifications,
            count_args,
            self._query_notifications,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return notifications_generator
        return list(notifications_generator)

    def _query_notifications(self,
                             skip: int,
                             first: int,
                             payload: dict,
                             fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_notifications = gql_notifications(
            fragment_builder(fields, Notification))
        result = self.auth.client.execute(_gql_notifications, payload)
        return format_result('data', result)

    @Compatible(['v2'])
    @typechecked
    def count_notifications(self,
                            has_been_seen: Optional[bool] = None,
                            user_id: Optional[str] = None) -> int:
        """Count the number of notifications.

        Args:
            has_been_seen: Filter on notifications that have been seen.
            user_id: Filter on the notifications of a specific user

        Returns:
            The number of notifications with the parameters provided
        """
        variables = {
            'where': {
                'user': {
                    'id': user_id,
                },
                'hasBeenSeen': has_been_seen,
            },
        }
        result = self.auth.client.execute(GQL_NOTIFICATIONS_COUNT, variables)
        count = format_result('data', result)
        return count



"""Project mutations."""

from json import dumps
from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from ...queries.asset import QueriesAsset
from ...queries.project import QueriesProject
from .helpers import verify_argument_ranges
from .queries import (GQL_APPEND_TO_ROLES,
                      GQL_CREATE_PROJECT,
                      GQL_DELETE_FROM_ROLES,
                      GQL_DELETE_PROJECT,
                      GQL_PROJECT_DELETE_ASYNCHRONOUSLY,
                      GQL_MAKE_PROJECT_PUBLIC,
                      GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER,
                      GQL_UPDATE_PROPERTIES_IN_PROJECT,
                      GQL_UPDATE_PROPERTIES_IN_ROLE)


class MutationsProject:
    """ Set of Project mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def append_to_roles(self, project_id: str, user_email: str, role: str = 'LABELER'):
        """Add a user to a project.

        !!! info
            If the user does not exist in your organization, he/she is invited and added
                both to your organization and project. This function can also be used to change
                the role of the user in the project.

        Args:
            project_id: Identifier of the project
            user_email: The email of the user.
                This email is used as the unique identifier of the user.
            role: One of {"ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"}.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.


        Examples:
            >>> kili.append_to_roles(project_id=project_id, user_email='john@doe.com')
        """
        variables = {
            'data': {'role': role,
                     'userEmail': user_email},
            'where': {'id': project_id}
        }
        result = self.auth.client.execute(GQL_APPEND_TO_ROLES, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_properties_in_project(self, project_id: str,
                                     consensus_mark: Optional[float] = None,
                                     consensus_tot_coverage: Optional[int] = None,
                                     description: Optional[str] = None,
                                     honeypot_mark: Optional[float] = None,
                                     instructions: Optional[str] = None,
                                     input_type: Optional[str] = None,
                                     json_interface: Optional[dict] = None,
                                     min_consensus_size: Optional[int] = None,
                                     number_of_assets: Optional[int] = None,
                                     number_of_assets_with_empty_labels: Optional[int] = None,
                                     number_of_remaining_assets: Optional[int] = None,
                                     number_of_reviewed_assets: Optional[int] = None,
                                     review_coverage: Optional[int] = None,
                                     should_relaunch_kpi_computation: Optional[bool] = None,
                                     title: Optional[str] = None,
                                     use_honeypot: Optional[bool] = None):
        """Update properties of a project.

        Args:
            project_id: Identifier of the project.
            consensus_mark: Should be between 0 and 1.
            consensus_tot_coverage: Should be between 0 and 100.
                It is the percentage of the dataset that will be annotated several times.
            description : Description of the project.
            honeypot_mark : Should be between 0 and 1
            instructions : Instructions of the project.
            interface_category: Always use 'IV2'.
            input_type: Currently, one of `AUDIO`, `FRAME`, `IMAGE`, `PDF`, `TEXT`, `VIDEO`, `NA`.
            json_interface: The json parameters of the project, see Edit your interface.
            min_consensus_size: Should be between 1 and 10
                Number of people that will annotate the same asset, for consensus computation.
            number_of_assets: Defaults to 0
            number_of_assets_with_empty_labels: Defaults to 0
            number_of_remaining_assets: Defaults to 0
            number_of_reviewed_assets: Defaults to 0
            review_coverage: Allow to set the percentage of assets
                that will be queued in the review interface.
                Should be between 0 and 100
            should_relaunch_kpi_computation: Technical field, added to indicate changes
                in honeypot or consensus settings
            title: Title of the project
            use_honeypot: Activate / Deactivate the use of honeypot in the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_project(project_id=project_id, title='New title')
        """
        verify_argument_ranges(consensus_tot_coverage,
                               min_consensus_size,
                               review_coverage)

        variables = {
            'consensusMark': consensus_mark,
            'consensusTotCoverage': consensus_tot_coverage,
            'description': description,
            'honeypotMark': honeypot_mark,
            'instructions': instructions,
            'inputType': input_type,
            'jsonInterface': dumps(json_interface) if json_interface is not None else None,
            'minConsensusSize': min_consensus_size,
            'numberOfAssets': number_of_assets,
            'numberOfAssetsWithSkippedLabels': number_of_assets_with_empty_labels,
            'numberOfRemainingAssets': number_of_remaining_assets,
            'numberOfReviewedAssets': number_of_reviewed_assets,
            'projectID': project_id,
            'reviewCoverage': review_coverage,
            'shouldRelaunchKpiComputation': should_relaunch_kpi_computation,
            'title': title,
            'useHoneyPot': use_honeypot
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)
        return format_result('data', result)

    @Compatible(endpoints=['v2'])
    @typechecked
    def create_project(self, input_type: str, json_interface: dict,
                       title: str, description: str = '',
                       project_type: Optional[str] = None):
        # pylint: disable=line-too-long
        """Create a project.

        Args:
            input_type : Currently, one of {AUDIO, IMAGE, PDF, TEXT, URL, VIDEO, NA}
            json_interface: The json parameters of the project, see Edit your interface.
            title : Title of the project
            description : Description of the project
            project_type:
                Currently, one of {
                    `IMAGE_CLASSIFICATION_SINGLE`,
                    `IMAGE_CLASSIFICATION_MULTI`,
                    `IMAGE_OBJECT_DETECTION_RECTANGLE`,
                    `IMAGE_OBJECT_DETECTION_POLYGON`,
                    `IMAGE_OBJECT_DETECTION_SEMANTIC`,
                    `OCR, PDF_CLASSIFICATION_SINGLE`,
                    `PDF_CLASSIFICATION_MULTI`,
                    `TEXT_CLASSIFICATION_SINGLE`,
                    `TEXT_CLASSIFICATION_MULTI`,
                    `TEXT_TRANSCRIPTION, TEXT_NER`,
                    `VIDEO_CLASSIFICATION_SINGLE`,
                    `VIDEO_FRAME_CLASSIFICATION`,
                    `VIDEO_FRAME_OBJECT_TRACKING`,
                    `SPEECH_TO_TEXT`
                }

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.create_project(input_type='IMAGE', json_interface=json_interface, title='Example')

        !!! example "Recipe"
            For more detailed examples on how to create projects,
                see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/create_project.ipynb).
        """
        variables = {
            'data': {'description': description,
                     'inputType': input_type,
                     'jsonInterface': dumps(json_interface),
                     'projectType': project_type,
                     'title': title}
        }
        result = self.auth.client.execute(GQL_CREATE_PROJECT, variables)
        return format_result('data', result)

    @Compatible(['v2'])
    @typechecked
    def make_project_public(self, project_id: str):
        """
        Make a project public.

        !!! warning
            This action is permanent and irreversible.

        Args:
            project_id: Identifier of the project

        Returns:
            The public token to provide in the public URL
        """
        variables = {'where': {'id': project_id}}
        result = self.auth.client.execute(GQL_MAKE_PROJECT_PUBLIC, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_properties_in_role(self, role_id: str,
                                  project_id: str, user_id: str, role: str):
        """Update properties of a role.

        !!! info
            To be able to change someone's role, you must be either of:

            - an admin of the project
            - a team manager of the project
            - an admin of the organization

        Args:
            role_id: Role identifier of the user. E.g. : 'to-be-deactivated'
            project_id: Identifier of the project
            user_id: The email or identifier of the user with updated role
            role: The new role.
                Possible choices are: `ADMIN`, `TEAM_MANAGER`, `REVIEWER`, `LABELER`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            'roleID': role_id,
            'projectID': project_id,
            'userID': user_id,
            'role': role
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_ROLE, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def delete_from_roles(self, role_id: str):
        """Delete users by their role_id.

        Args:
            role_id : Identifier of the project user (not the ID of the user)

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {'where': {'id': role_id}}
        result = self.auth.client.execute(GQL_DELETE_FROM_ROLES, variables)
        return format_result('data', result)

    @Compatible(['v2'])
    @typechecked
    def update_properties_in_project_user(self, project_user_id: str,
                                          consensus_mark: Optional[float] = None,
                                          honeypot_mark: Optional[float] = None,
                                          number_of_labeled_assets: Optional[int] = None,
                                          starred: Optional[bool] = None,
                                          total_duration: Optional[int] = None):
        """
        Update properties of a project-user tuple

        Args:
            project_user_id : Identifier of the project user
            consensus_mark: Should be between 0 and 1.
            honeypot_mark: Should be between 0 and 1.
            number_of_labeled_assets: Number of assets the user labeled in the project.
            starred: Whether to star the project in the project list.
            total_duration: Total time the user spent in the project.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> for project_user in project_users:
            ...     kili.update_properties_in_project_user(
                        project_user_id=project_user['id'],
                        honeypot_mark=0)
        """
        variables = {
            'consensusMark': consensus_mark,
            'honeypotMark': honeypot_mark,
            'numberOfLabeledAssets': number_of_labeled_assets,
            'projectUserID': project_user_id,
            'starred': starred,
            'totalDuration': total_duration,
        }
        result = self.auth.client.execute(
            GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def internal_delete_project(self, project_id: str):
        """Delete project permanently.
        WARNING: This resolver is for internal use by Kili Technology only.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {'projectID': project_id}
        result = self.auth.client.execute(GQL_DELETE_PROJECT, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def delete_project(self, project_id: str):
        """
        Delete a project permanently.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {'where': {'id': project_id}}
        result = self.auth.client.execute(
            GQL_PROJECT_DELETE_ASYNCHRONOUSLY, variables)
        return format_result('data', result)



"""Project queries."""

from typing import Generator, List, Optional, Union
import warnings
from typeguard import typechecked


from ...helpers import Compatible, deprecate, format_result, fragment_builder
from .queries import gql_projects, GQL_PROJECTS_COUNT
from ...types import Project
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesProject:
    """Set of Project queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v1', 'v2'])
    @typechecked
    def projects(self,
                 project_id: Optional[str] = None,
                 search_query: Optional[str] = None,
                 should_relaunch_kpi_computation: Optional[bool] = None,
                 updated_at_gte: Optional[str] = None,
                 updated_at_lte: Optional[str] = None,
                 skip: int = 0,
                 fields: List[str] = [
                     'consensusTotCoverage',
                     'id',
                     'inputType',
                     'jsonInterface',
                     'minConsensusSize',
                     'reviewCoverage',
                     'roles.id',
                     'roles.role',
                     'roles.user.email',
                     'roles.user.id',
                     'title'],
                 first: int = 100,
                 disable_tqdm: bool = False,
                 as_generator: bool = False) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of projects that match a set of criteria.

        Args:
            project_id: Select a specific project through its project_id.
            search_query: Returned projects with a title or a description matching this string.
            should_relaunch_kpi_computation : Technical field, added to indicate changes in honeypot or consensus settings.
            updated_at_gte: Returned projects should have a label whose update date is greater or equal
                to this date.
            updated_at_lte: Returned projects should have a label whose update date is lower or equal to this date.
            skip: Number of projects to skip (they are ordered by their creation).
            fields: All the fields to request among the possible fields for the projects.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#project) for all possible fields.
            first: Maximum number of projects to return.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the projects is returned.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> # List all my projects
            >>> kili.projects()
        """

        saved_args = locals()
        count_args = {
            k: v
            for (k, v) in saved_args.items() if k in [
                'project_id',
                'search_query',
                'should_relaunch_kpi_computation',
                'updated_at_gte',
                'updated_at_lte'
            ]
        }
        disable_tqdm = disable_tqdm or as_generator

        payload_query = {
            'where': {
                'id': project_id,
                'searchQuery': search_query,
                'shouldRelaunchKpiComputation': should_relaunch_kpi_computation,
                'updatedAtGte': updated_at_gte,
                'updatedAtLte': updated_at_lte,
            },
        }

        projects_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_projects,
            count_args,
            self._query_projects,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return projects_generator
        return list(projects_generator)

    def _query_projects(self,
                        skip: int,
                        first: int,
                        payload: dict,
                        fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_projects = gql_projects(fragment_builder(fields, Project))
        result = self.auth.client.execute(_gql_projects, payload)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def count_projects(
            self,
            project_id: Optional[str] = None,
            search_query: Optional[str] = None,
            should_relaunch_kpi_computation: Optional[bool] = None,
            updated_at_gte: Optional[str] = None,
            updated_at_lte: Optional[str] = None) -> int:
        """
        Counts the number of projects with a search_query

        Args:
            project_id: Select a specific project through its project_id.
            search_query: Returned projects have a title or a description that matches this string.
            should_relaunch_kpi_computation : Technical field, added to indicate changes in honeypot
                or consensus settings
            updated_at_gte: Returned projects should have a label
                whose update date is greater
                or equal to this date.
            updated_at_lte: Returned projects should have a label
                whose update date is lower or equal to this date.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of projects with the parameters provided
        """
        variables = {
            'where': {
                'id': project_id,
                'searchQuery': search_query,
                'shouldRelaunchKpiComputation': should_relaunch_kpi_computation,
                'updatedAtGte': updated_at_gte,
                'updatedAtLte': updated_at_lte,
            }
        }
        result = self.auth.client.execute(GQL_PROJECTS_COUNT, variables)
        count = format_result('data', result)
        return count



"""Project user queries."""

from typing import Generator, List, Optional, Union
import warnings

from typeguard import typechecked


from ...helpers import Compatible, deprecate, format_result, fragment_builder
from .queries import gql_project_users, GQL_PROJECT_USERS_COUNT
from ...types import ProjectUser
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesProjectUser:
    """Set of ProjectUser queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value,invalid-name
    @Compatible(['v1', 'v2'])
    @typechecked
    def project_users(self,
                      email: Optional[str] = None,
                      id: Optional[str] = None,  # pylint: disable=redefined-builtin
                      organization_id: Optional[str] = None,
                      project_id: Optional[str] = None,
                      fields: List[str] = ['activated', 'id', 'role',
                                           'starred', 'user.email', 'user.id'],
                      first: int = 100,
                      skip: int = 0,
                      disable_tqdm: bool = False,
                      as_generator: bool = False) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Return project users (possibly with their KPIs) that match a set of criteria


        Args:
            email: Email of the user
            organization_id: Identifier of the user's organization
            project_id: Identifier of the project
            fields: All the fields to request among the possible fields for the projectUsers
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#projectuser) for all possible fields.
            first: Maximum number of users to return
            skip: Number of project users to skip
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the project users is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            ```
            # Retrieve consensus marks of all users in project
            >>> kili.project_users(project_id=project_id, fields=['consensusMark', 'user.email'])
            ```
        """

        count_args = {"email": email,
                      "id": id,
                      "organization_id": organization_id,
                      "project_id": project_id}
        disable_tqdm = disable_tqdm or as_generator
        payload_query = {
            'where': {
                'id': id,
                'project': {
                    'id': project_id,
                },
                'user': {
                    'email': email,
                    'organization': {
                        'id': organization_id,
                    }
                },
            }
        }

        project_users_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_project_users,
            count_args,
            self._query_project_users,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return project_users_generator
        return list(project_users_generator)

    def _query_project_users(self,
                             skip: int,
                             first: int,
                             payload: dict,
                             fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_project_users = gql_project_users(
            fragment_builder(fields, ProjectUser))
        result = self.auth.client.execute(_gql_project_users, payload)
        return format_result('data', result)

    # pylint: disable=invalid-name
    @typechecked
    def count_project_users(
            self,
            email: Optional[str] = None,
            id: Optional[str] = None,  # pylint: disable=redefined-builtin
            organization_id: Optional[str] = None,
            project_id: Optional[str] = None) -> int:
        """
        Counts the number of projects and their users that match a set of criteria

        Args:
            email: Email of the user
            organization_id: Identifier of the user's organization
            project_id: Identifier of the project

        Returns:
            The number of project users with the parameters provided
        """
        variables = {
            'where': {
                'id': id,
                'project': {
                    'id': project_id,
                },
                'user': {
                    'email': email,
                    'organization': {
                        'id': organization_id,
                    }
                },
            }
        }
        result = self.auth.client.execute(GQL_PROJECT_USERS_COUNT, variables)
        count = format_result('data', result)
        return count



"""
This script permits to initialize the Kili Python SDK client.
"""
import os

from kili.mutations.api_key import MutationsApiKey
from kili.mutations.asset import MutationsAsset
from kili.mutations.label import MutationsLabel
from kili.mutations.notification import MutationsNotification
from kili.mutations.organization import MutationsOrganization
from kili.mutations.project import MutationsProject
from kili.mutations.project_version import MutationsProjectVersion
from kili.mutations.user import MutationsUser
from kili.queries.api_key import QueriesApiKey
from kili.queries.asset import QueriesAsset
from kili.queries.issue import QueriesIssue
from kili.queries.label import QueriesLabel
from kili.queries.lock import QueriesLock
from kili.queries.organization import QueriesOrganization
from kili.queries.notification import QueriesNotification
from kili.queries.project import QueriesProject
from kili.queries.project_user import QueriesProjectUser
from kili.queries.project_version import QueriesProjectVersion
from kili.queries.user import QueriesUser
from kili.subscriptions.label import SubscriptionsLabel


from kili.authentication import KiliAuth


class Kili(  # pylint: disable=too-many-ancestors
        MutationsApiKey,
        MutationsAsset,
        MutationsLabel,
        MutationsNotification,
        MutationsOrganization,
        MutationsProject,
        MutationsProjectVersion,
        MutationsUser,
        QueriesApiKey,
        QueriesAsset,
        QueriesIssue,
        QueriesLabel,
        QueriesLock,
        QueriesOrganization,
        QueriesNotification,
        QueriesProject,
        QueriesProjectUser,
        QueriesProjectVersion,
        QueriesUser,
        SubscriptionsLabel):
    """
    Kili Client.
    """

    def __init__(self, api_key=os.getenv('KILI_API_KEY'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/v2/graphql',
                 verify=True):
        """
        Args:
            api_key: User API key generated
                from https://cloud.kili-technology.com/label/my-account/api-key
            api_endpoint: Recipient of the HTTP operation
            verify: Verify certificate. Set to False on local deployment without SSL.

        Returns:
            Object container your API session

        Examples:
            list:
                - your assets with: `kili.assets()`
                - your labels with: `kili.labels()`
                - your projects with: `kili.projects()`
        """
        self.auth = KiliAuth(
            api_key=api_key, api_endpoint=api_endpoint, verify=verify)
        super().__init__(self.auth)



"""
Fragments of label mutations
"""

LABEL_FRAGMENT = '''
id
author {
  id
  email
}
labelType
jsonResponse
createdAt
secondsToLabel
totalSecondsToLabel
honeypotMark
'''


LABEL_FRAGMENT_ID = '''
id
'''



"""
Queries of project mutations
"""

from .fragments import (PROJECT_FRAGMENT, PROJECT_FRAGMENT_ID,
                        PROJECT_FRAGMENT_PUBLIC_TOKEN, ROLE_FRAGMENT)


GQL_APPEND_TO_ROLES = f'''
mutation($data: AppendToRolesData!, $where: ProjectWhere!) {{
  data: appendToRoles(
    data: $data
    where: $where
  ) {{
      {PROJECT_FRAGMENT}
  }}
}}
'''

GQL_UPDATE_PROPERTIES_IN_PROJECT = f'''
mutation(
    $consensusMark: Float
    $consensusTotCoverage: Int
    $description: String
    $honeypotMark: Float
    $instructions: String
    $inputType: InputType
    $jsonInterface: String
    $minConsensusSize: Int
    $numberOfAssets: Int
    $numberOfAssetsWithSkippedLabels: Int
    $numberOfRemainingAssets: Int
    $numberOfReviewedAssets: Int
    $projectID: ID!
    $reviewCoverage: Int
    $title: String
    $useHoneyPot: Boolean
) {{
  data: updatePropertiesInProject(
    where: {{
      id: $projectID
    }},
    data: {{
      consensusMark: $consensusMark
      consensusTotCoverage: $consensusTotCoverage
      description: $description
      honeypotMark: $honeypotMark
      instructions: $instructions
      inputType: $inputType
      jsonInterface: $jsonInterface
      minConsensusSize: $minConsensusSize
      numberOfAssets: $numberOfAssets
      numberOfAssetsWithSkippedLabels: $numberOfAssetsWithSkippedLabels
      numberOfRemainingAssets: $numberOfRemainingAssets
      numberOfReviewedAssets: $numberOfReviewedAssets
      reviewCoverage: $reviewCoverage
      title: $title
      useHoneyPot: $useHoneyPot
    }}
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_CREATE_PROJECT = f'''
mutation(
    $data: CreateProjectData!
  ) {{
  data: createProject(
    data: $data
  ) {{
      {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_MAKE_PROJECT_PUBLIC = f'''
mutation($where: ProjectWhere!) {{
  data: makeProjectPublic(where: $where) {{
    {PROJECT_FRAGMENT_PUBLIC_TOKEN}
  }}
}}
'''

GQL_UPDATE_PROPERTIES_IN_ROLE = f'''
mutation(
    $roleID: ID!
    $projectID: ID!
    $userID: ID!
    $role: ProjectRole!
) {{
  data: updatePropertiesInRole(
    where: {{id: $roleID}}
    data: {{
      projectID: $projectID
      userID: $userID
      role: $role
    }}) {{
      {ROLE_FRAGMENT}
  }}
}}
'''

GQL_DELETE_FROM_ROLES = f'''
mutation($where: ProjectUserWhere!) {{
  data: deleteFromRoles(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER = f'''
mutation(
    $projectUserID: ID!
    $totalDuration: Int
    $numberOfLabeledAssets: Int
    $starred: Boolean
    $consensusMark: Float
    $honeypotMark: Float
) {{
  data: updatePropertiesInProjectUser(
    where: {{id: $projectUserID}},
    data: {{
      totalDuration: $totalDuration
      numberOfLabeledAssets: $numberOfLabeledAssets
      starred: $starred
      consensusMark: $consensusMark
      honeypotMark: $honeypotMark
    }}
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_DELETE_PROJECT = f'''
mutation($projectID: ID!) {{
  data: deleteProject(where: {{
      id: $projectID
    }}
    ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_PROJECT_DELETE_ASYNCHRONOUSLY = '''
mutation($where: ProjectWhere!) {
  data: deleteProjectAsynchronously(where: $where)
}
'''



"""
Label mutations
"""

from json import dumps
from typing import List, Optional
import warnings
from typeguard import typechecked


from ...helpers import Compatible, format_result, infer_id_from_external_id
from .queries import (GQL_APPEND_TO_LABELS, GQL_CREATE_HONEYPOT,
                      GQL_CREATE_PREDICTIONS,
                      GQL_UPDATE_PROPERTIES_IN_LABEL)
from ...orm import Label
from ...utils.pagination import _mutate_from_paginated_call


class MutationsLabel:
    """Set of Label mutations."""
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initializes the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def create_predictions(
            self,
            project_id: str,
            external_id_array: List[str],
            model_name_array: List[str],
            json_response_array: List[dict]):
        # pylint: disable=line-too-long
        """Create predictions for specific assets.

        Args:
            project_id: Identifier of the project
            external_id_array: The external identifiers of the assets for which we want to add predictions
            model_name_array: In case you want to precise from which model the label originated
            json_response_array: The predictions are given here. For examples,
                see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_predictions.ipynb).

        Returns:
            A result object which indicates if the mutation was successful, or an error message.

        !!! example "Recipe"
            For more detailed examples on how to create predictions, see [the recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/import_predictions.ipynb).
        """
        assert len(external_id_array) == len(
            json_response_array), "IDs list and predictions list should have the same length"
        assert len(external_id_array) == len(
            model_name_array), "IDs list and model names list should have the same length"
        if len(external_id_array) == 0:
            warnings.warn("Empty IDs and prediction list")

        properties_to_batch = {'external_id_array': external_id_array,
                               'model_name_array': model_name_array,
                               'json_response_array': json_response_array}

        def generate_variables(batch):
            return {
                'data': {'modelNameArray': batch['model_name_array'],
                         'jsonResponseArray': [dumps(elem) for elem in batch['json_response_array']]},
                'where': {'externalIdStrictlyIn': batch['external_id_array'], 'project': {'id': project_id}}
            }

        results = _mutate_from_paginated_call(
            self, properties_to_batch, generate_variables, GQL_CREATE_PREDICTIONS)
        return format_result('data', results[0], Label)

    @Compatible(['v1', 'v2'])
    @typechecked
    def append_to_labels(self, json_response: dict, author_id: Optional[str] = None,
                         label_asset_external_id: Optional[str] = None,
                         label_asset_id: Optional[str] = None, label_type: str = 'DEFAULT',
                         project_id: Optional[str] = None, seconds_to_label: Optional[int] = 0,
                         skipped: Optional[bool] = False):
        """Append a label to an asset.

        Args:
            json_response: Label is given here
            author_id: ID of the author of the label
            label_asset_external_id: External identifier of the asset
            label_asset_id: Identifier of the asset
            project_id: Identifier of the project
            label_type: Can be one of `AUTOSAVE`, `DEFAULT`, `PREDICTION` or `REVIEW`
            seconds_to_label: Time to create the label
            skipped: Describe if the label is skipped or not

        !!! warning
            Either provide `label_asset_id` or `label_asset_external_id` and `project_id`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(label_asset_id=asset_id, json_response={...})

        """

        if author_id is None:
            author_id = self.auth.user_id
        label_asset_id = infer_id_from_external_id(
            self, label_asset_id, label_asset_external_id, project_id)
        variables = {
            'data': {'authorID': author_id,
                     'jsonResponse': dumps(json_response),
                     'labelType': label_type,
                     'secondsToLabel': seconds_to_label,
                     'skipped': skipped},
            'where': {'id': label_asset_id}
        }
        result = self.auth.client.execute(GQL_APPEND_TO_LABELS, variables)
        return format_result('data', result, Label)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_properties_in_label(self,
                                   label_id: str,
                                   seconds_to_label: Optional[int] = None,
                                   model_name: Optional[str] = None,
                                   json_response: Optional[dict] = None):
        """Update properties of a label.

        Args:
            label_id: Identifier of the label
            seconds_to_label: Time to create the label
            model_name: Name of the model
            json_response: The label is given here

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_label(label_id=label_id, json_response={...})
        """
        formatted_json_response = None if json_response is None else dumps(
            json_response)
        variables = {
            'labelID': label_id,
            'secondsToLabel': seconds_to_label,
            'modelName': model_name,
            'jsonResponse': formatted_json_response
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_LABEL, variables)
        return format_result('data', result, Label)

    @Compatible(['v1', 'v2'])
    @typechecked
    def create_honeypot(self, json_response: dict, asset_external_id: Optional[str] = None,
                        asset_id: Optional[str] = None, project_id: Optional[str] = None):
        """Create honeypot for an asset.

        !!! info
            Uses the given `json_response` to create a `REVIEW` label.
            This enables Kili to compute a`honeypotMark`,
            which measures the similarity between this label and other labels.

        Args:
            json_response: The JSON response of the honeypot label of the asset
            asset_id: Identifier of the asset
                Either provide asset_id or asset_external_id and project_id
            asset_external_id: External identifier of the asset
                Either provide asset_id or asset_external_id and project_id
            project_id: Identifier of the project
                Either provide asset_id or asset_external_id and project_id

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        asset_id = infer_id_from_external_id(
            self, asset_id, asset_external_id, project_id)

        variables = {
            'data': {'jsonResponse': dumps(json_response)},
            'where': {'id': asset_id}
        }
        result = self.auth.client.execute(GQL_CREATE_HONEYPOT, variables)
        return format_result('data', result, Label)



"""
Queries of project version queries
"""


def gql_project_version(fragment):
    """
    Return the GraphQL projectVersion query
    """
    return f'''
query ($where: ProjectVersionWhere!, $first: PageSize!, $skip: Int!) {{
  data: projectVersions(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
'''


GQL_PROJECT_VERSION_COUNT = '''
query($where: ProjectVersionWhere!) {
  data: countProjectVersions(where: $where)
}
'''



"""Label subscription."""

from dataclasses import dataclass
from typing import Any, Callable

from typeguard import typechecked

from .subscriptions import GQL_LABEL_CREATED_OR_UPDATED
from ...graphql_client import SubscriptionGraphQLClient


@dataclass
class SubscriptionsLabel:
    """Set of Label subscriptions."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def label_created_or_updated(self, project_id: str, callback: Callable[[str, str], None]):
        # pylint: disable=line-too-long
        """
        Subscribe a callback to a project, which is executed when a label is created or updated.

        Args:
            project_id: Identifier of the project
            callback: This function takes as input the id of the asset and its content.

        Returns:
            A subscription client

        !!! example "Recipe"
            For more detailed examples on how to use Webhooks,
            See [the related recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/webhooks.ipynb)
        """
        ws_endpoint = self.auth.client.endpoint.replace('http', 'ws')
        websocket = SubscriptionGraphQLClient(ws_endpoint)
        headers = {'Accept': 'application/json',
                   'Content-Type': 'application/json'}
        authorization = f'{self.auth.client.token}'
        headers['Authorization'] = authorization
        variables = {'projectID': project_id}
        websocket.subscribe(
            GQL_LABEL_CREATED_OR_UPDATED,
            variables=variables,
            callback=callback,
            headers=headers,
            authorization=authorization)
        return websocket



"""
Api key mutations
"""

from typeguard import typechecked
from ...helpers import Compatible, deprecate, format_result

from .queries import GQL_APPEND_TO_API_KEYS


class MutationsApiKey:  # pylint: disable=too-few-public-methods
    """Set of User mutations."""
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def append_to_api_keys(self, api_key: str, name: str):
        """Create an api key to connect to the API.

        Args:
            api_key: A new api key to connect with
            name: A name used to describe the api key.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            'data': {'key': api_key,
                     'name': name},
            'where': {'email': self.auth.user_email}
        }
        result = self.auth.client.execute(GQL_APPEND_TO_API_KEYS, variables)
        return format_result('data', result)



"""
Api key queries
"""

from typing import Generator, List, Optional, Union
import warnings
from typeguard import typechecked

from ...utils.pagination import row_generator_from_paginated_calls
from ...helpers import Compatible, format_result, fragment_builder
from .queries import gql_api_keys, GQL_API_KEYS_COUNT
from ...types import ApiKey as ApiKeyType


class QueriesApiKey:
    """
    Set of ApiKey queries
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v2'])
    @typechecked
    def api_keys(self, api_key_id: Optional[str] = None, user_id: Optional[str] = None,
                 api_key: Optional[str] = None, skip: int = 0,
                 fields: List[str] = ['id', 'name', 'createdAt', 'revoked'],
                 first: Optional[int] = 100,
                 disable_tqdm: bool = False,
                 as_generator: bool = False) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of API keys that match a set of constraints.

        !!! info
            You can only query your own API keys

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the API key.
            skip: Number of assets to skip (they are ordered by their date of creation, first to last).
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#apikey) for all possible fields.
            first: Maximum number of API keys to return.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the API key is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.


        Examples:
            >>> kili.api_keys(user_id=user_id)
            >>> kili.api_keys(api_key=api_key)
            >>> kili.api_keys(api_key=api_key, as_generator=False)
        """

        saved_args = locals()
        count_args = {
            k: v
            for (k, v) in saved_args.items()
            if k
            in [
                'user_id',
                'api_key_id',
                'api_key'
            ]
        }
        disable_tqdm = disable_tqdm or as_generator

        payload_query = {
            'where': {
                'user': {
                    'id': user_id,
                    'apiKey': api_key
                },
                'id': api_key_id,
            },
        }

        api_keys_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_api_keys,
            count_args,
            self._query_api_keys,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return api_keys_generator
        return list(api_keys_generator)

    def _query_api_keys(self,
                        skip: int,
                        first: int,
                        payload: dict,
                        fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_api_keys = gql_api_keys(fragment_builder(fields, ApiKeyType))
        result = self.auth.client.execute(_gql_api_keys, payload)
        return format_result('data', result)

    @Compatible(['v2'])
    @typechecked
    def count_api_keys(self, api_key_id: Optional[str] = None, user_id: Optional[str] = None,
                       api_key: Optional[str] = None) -> int:
        """Count and return the number of api keys with the given constraints.

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the api key.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.count_api_keys(user_id=user_id)
            3
            >>> kili.count_api_keys(api_key=api_key)
            1
        """
        variables = {
            'where': {
                'user': {
                    'id': user_id,
                    'apiKey': api_key
                },
                'id': api_key_id,
            },
        }
        result = self.auth.client.execute(GQL_API_KEYS_COUNT, variables)
        count = format_result('data', result)
        return count



"""Tests for utils module"""

from kili.utils.pagination import (
    batch_iterator_builder, batch_object_builder, row_generator_from_paginated_calls)
from .utils import mocked_count_method, mocked_query_method

TEST_CASES = [
    {
        "case": "AAU, When I query objects with first=50, I get the first 50 objects",
        "args": {"first": 50},
        "expected_result": ({"id": i} for i in range(50)),
    },
    {
        "case": "AAU, When I query objects with first=200, I get the first 200 objects",
        "args": {"first": 200},
        "expected_result": ({"id": i} for i in range(200)),
    },
    {
        "case": "AAU, When I query objects with first=50 and skip=20, I skip the 20"
        " first objects I get the 50 objects following",
        "args": {"first": 50, "skip": 20},
        "expected_result": ({"id": i} for i in range(20, 70)),
    },
    {
        "case": "AAU, When I query objects with first=50 and skip=20, I skip the 20"
        "first objects I get the 200 following objects",
        "args": {"first": 200, "skip": 20},
        "expected_result": ({"id": i} for i in range(20, 220)),
    },
    {
        "case": "AAU, When I query objects with first=25100, I get the first 25100 objects",
        "args": {"first": 25100},
        "expected_result": ({"id": i} for i in range(25100)),
    },
    {
        "case": "AAU, When I query objects with first=20 and disable_tqdm, I get the first 2O objects",
        "args": {"first": 20, "disable_tqdm": True},
        "expected_result": ({"id": i} for i in range(20)),
    },
    {
        "case": "AAU, When I query objects with first=20, skip=20 and disable_tqdm, I  skip the 20"
        "first objects and I get the 20 following objects",
        "args": {"first": 20, "skip": 20, "disable_tqdm": True},
        "expected_result": ({"id": i} for i in range(20, 40)),
    },
    {
        "case": "AAU, When I query objects with first=0, I get no objects",
        "args": {"first": 0},
        "expected_result": iter(()),
    },
]


def test_row_generator_from_paginated_calls():
    """
        Simulates a count query result by returning a list of ids
    """
    for test_case in TEST_CASES:
        case_name = test_case["case"]
        expected = test_case["expected_result"]
        skip = test_case["args"].get('skip', 0)
        first = test_case["args"].get('first')
        disable_tqdm = test_case["args"].get('disable_tqdm', False)

        actual = row_generator_from_paginated_calls(
            skip,
            first,
            mocked_count_method,
            {},
            mocked_query_method,
            {first, skip},
            [],
            disable_tqdm)
        assert all(a == b for a, b in zip(actual, expected)), \
            f"Test case \"{case_name}\" failed"
        assert type(actual).__name__ == "generator", \
            f"Test case \"{case_name}\" failed"


def test_batch_iterator_builder():
    """Test batch iterator builder."""
    TEST_CASE = [{
        'case': 'When the size of the iterable is not a multiple of the batch_size, I see a success',
        'batch_size': 3,
        'iterable': [i for i in range(7)],
        'expected_result': (i for i in [[0, 1, 2], [3, 4, 5], [6]])
    }]
    for test_case in TEST_CASE:
        actual = batch_iterator_builder(
            test_case['iterable'], batch_size=test_case['batch_size'])
        expected = test_case['expected_result']
        case_name = test_case['case']
        assert all(a == b for a, b in zip(actual, expected)
                   ), f"Test case \"{case_name}\" failed"


def test_batch_object_builder():
    """Test batch iterator builder for several arrays in the same time."""
    array1 = [i for i in range(10)]
    array2 = [i for i in range(10, 20)]
    array3 = [i for i in range(20, 30)]
    TEST_CASE = [{
        'case': 'When I have one property I see that it returns paginated batches',
        'batch_size': 3,
        'properties_to_batch': {'a1': array1},
        'expected_result': (i for i in [{'a1': [0, 1, 2]},
                                        {'a1': [3, 4, 5]},
                                        {'a1': [6, 7, 8]},
                                        {'a1': [9]}])
    },
        {
        'case': 'When I have several iterables I see it creates a common batch',
        'batch_size': 4,
        'properties_to_batch': {'a1': array1, 'a2': array2, 'a3': array3},
        'expected_result': (i for i in [{'a1': [0, 1, 2, 3], 'a2':[10, 11, 12, 13],
                                         'a3':[20, 21, 22, 23]},
                                        {'a1': [4, 5, 6, 7], 'a2':[14, 15, 16, 17],
                                         'a3':[24, 25, 26, 27]},
                                        {'a1': [8, 9], 'a2':[18, 19], 'a3':[28, 29]}])
    },
        {
        'case': 'When I have None arrays, I see that it stays None in the batches',
        'batch_size': 4,
        'properties_to_batch': {'a1': array1, 'a2': None, 'a3': array2, 'a4': None},
        'expected_result': (i for i in [{'a1': [0, 1, 2, 3], 'a2':None,
                                         'a3':[10, 11, 12, 13], 'a4':None},
                                        {'a1': [4, 5, 6, 7], 'a2':None,
                                         'a3':[14, 15, 16, 17], 'a4':None},
                                        {'a1': [8, 9], 'a2':None, 'a3':[18, 19], 'a4':None}])
    },
        {
        'case': 'When I have only None, I see it return one batch of None',
                'batch_size': 2,
                'properties_to_batch': {'a1': None, 'a2': None, 'a3': None},
                'expected_result': (i for i in [{'a1': None, 'a2': None, 'a3': None}])
    }
    ]
    for test_case in TEST_CASE:
        actual = batch_object_builder(
            test_case['properties_to_batch'], test_case['batch_size'])
        expected = test_case['expected_result']
        case_name = test_case['case']
        assert all(a == b for a, b in zip(actual, expected)
                   ), f"Test case \"{case_name}\" failed"



"""
Helpers for GraphQL Queries and Mutations
"""

from typing import Optional
import base64
import functools
from json import dumps, loads
import re
import warnings
import mimetypes
import pyparsing as pp

import requests


class Compatible():
    """
    Compatibility of Kili Python SDK version with Kili API version
    """

    # pylint: disable=dangerous-default-value
    def __init__(self, endpoints=['v1']):
        self.endpoints = endpoints
        self.version_extractor = re.compile(r'\/v\d+/')
        self.address_extractor = re.compile(r':400\d+/')

    def client_is_compatible(self, endpoint: str):
        """
        Checks if client is compatible with Kili API

        Args:
            endpoint: the Kili API endpoint
        """
        version_matched = self.version_extractor.search(endpoint)
        address_matched = self.address_extractor.search(endpoint)
        if not version_matched and not address_matched:
            return False
        if address_matched:
            version = 'v1' if address_matched.group() == ':4000/' else 'v2'
        if version_matched:
            version = 'v1' if version_matched.group() == '/v1/' else 'v2'
        return version in self.endpoints

    def __call__(self, resolver, *args, **kwargs):
        @functools.wraps(resolver)
        def checked_resolver(*args, **kwargs):
            try:
                client_endpoint = args[0].auth.client.endpoint
            except Exception as exception:
                raise ValueError(
                    'Cannot find client endpoint from resolver'
                    f' {resolver.__name__} with arguments {args}') from exception
            if self.client_is_compatible(client_endpoint):
                return resolver(*args, **kwargs)
            raise EndpointCompatibilityError(
                resolver.__name__, client_endpoint)
        return checked_resolver


class EndpointCompatibilityError(Exception):
    """
    EndpointCompatibilityError
    """

    def __init__(self, resolver, endpoint):
        super().__init__(
            f'Resolver {resolver} is not compatible with the following endpoint : {endpoint}')


class GraphQLError(Exception):
    """
    GraphQLError
    """

    def __init__(self, mutation, error, batch_number=None):
        if batch_number is None:
            super().__init__(
                f'Mutation "{mutation}" failed with error: "{error}"')
        else:
            super().__init__(
                f'Mutation "{mutation}" failed from index {100*batch_number} with error: "{error}"')


def format_result(name, result, _object=None):
    """
    Formats the result of the GraphQL queries.

    Args:
        name: name of the field to extract, usually data
        result: query result to parse
    """
    if 'errors' in result:
        raise GraphQLError(name, result['errors'])
    formatted_json = format_json(result['data'][name])
    if _object is None:
        return formatted_json
    if isinstance(formatted_json, list):
        return [_object(element) for element in formatted_json]
    return _object(formatted_json)


def content_escape(content):
    """
    Escapes the content

    Args:
        content: string to escape
    """
    return content.replace('\\', '\\\\').replace('\n', '\\n').replace('"', '\\"')


def get_data_type(path):
    """
    Get the data type, either image/png or application/pdf

    Args:
        path: path of the file
    """
    mime_type, _ = mimetypes.guess_type(path.lower())
    return mime_type if mime_type else ''


def encode_base64(path):
    """
    Encode a file in base 64

    Args:
        path: path of the file
    """
    data_type = get_data_type(path)
    with open(path, 'rb') as image_file:
        return f'data:{data_type};base64,' + \
            base64.b64encode(image_file.read()).decode('ascii')


def is_url(path):
    """
    Check if the path is a url or something else

    Args:
        path: path of the file
    """
    return isinstance(path, str) and re.match(r'^(http://|https://)', path.lower())


def format_json_dict(result):
    """
    Formats the dict part of a json return by a GraphQL query into a python object

    Args:
        result: result of a GraphQL query
    """
    for key, value in result.items():
        if key in ['jsonInterface', 'jsonMetadata', 'jsonResponse']:
            if (value == '' or value is None) \
                    and not (is_url(value) and key == 'jsonInterface'):
                result[key] = {}
            elif isinstance(value, str):
                try:
                    if is_url(value):
                        result[key] = requests.get(value).json()
                    else:
                        result[key] = loads(value)
                except Exception as exception:
                    raise ValueError(
                        'Json Metadata / json response /'
                        ' json interface should be valid jsons') from exception
        else:
            result[key] = format_json(value)
    return result


def format_json(result):
    """
    Formats the json return by a GraphQL query into a python object

    Args:
        result: result of a GraphQL query
    """
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem) for elem in result]
    if isinstance(result, dict):
        return format_json_dict(result)
    return result


def fragment_builder(fields, type_of_fields):
    """
    Builds a GraphQL fragment for a list of fields to query

    Args:
        fields
        type_of_fields
    """
    fragment = ''
    subfields = [field.split('.', 1) for field in fields if '.' in field]
    if subfields:
        for subquery in {subfield[0] for subfield in subfields}:
            type_of_fields_subquery = getattr(type_of_fields, subquery)
            try:
                if issubclass(type_of_fields_subquery, object):
                    fields_subquery = [subfield[1]
                                       for subfield in subfields if subfield[0] == subquery]
                    new_fragment = fragment_builder(
                        fields_subquery, type_of_fields_subquery)
                    fragment += f' {subquery}{{{new_fragment}}}'
            except ValueError:
                print(f'{subquery} must be a valid subquery field')
        fields = [field for field in fields if '.' not in field]
    for field in fields:
        try:
            getattr(type_of_fields, field)
        except ValueError:
            print(f'{field} must be an instance of {type_of_fields}')
        if isinstance(field, str):
            fragment += f' {field}'
        else:
            raise Exception('Please provide the fields to query as strings')
    return fragment


def deprecate(
        msg: Optional[str] = None,
        removed_in: Optional[str] = None,
        _type=DeprecationWarning):
    """
    Decorator factory that tag a deprecated function.
    - To deprecated the whole function, you can give a message at the decorator level.
    - For more sharp condition on the warning message, integrate this warning inside the function
    but still tag the function with this decorator, without giving a message argument

    Args:
        msg: string message that will be displayed whenever the function is called
        removed_in: string version in the format "Major.Minor"
            in which the deprecation element has to be removed
        type: DeprecationWarning by default
    """
    def decorator(func):
        if removed_in:
            if len(removed_in.split('.')) != 2:
                raise ValueError(
                    f'"removed_in" argument in deprecate wrapper of the function {func.__name__}'
                    'should have the format "Major.Minor"')
            func.removed_in = removed_in

        @ functools.wraps(func)
        def wrapper(*args, **kwargs):
            if msg:
                warnings.warn(msg, _type, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def format_metadata(metadata):
    """
    Formats metadata

    Args:
        metadata: a python object
    """
    if metadata is None:
        return metadata
    if isinstance(metadata, str):
        return metadata
    if isinstance(metadata, (dict, list)):
        return dumps(metadata)
    raise Exception(
        f'Metadata {metadata} of type {type(metadata)} must either be None,'
        ' a string a list or a dict.')


def convert_to_list_of_none(array, length):
    """
    Turns a value in a list of length length

    Args:
        array
        length
    """
    if isinstance(array, list):
        if len(array) != length:
            raise Exception(f'array should have length {length}')
        return array
    return [None] * length


def is_none_or_empty(_object):
    """
    Tests if an object is none or empty

    Args:
        object: a python object
    """
    object_is_empty = isinstance(_object, list) and len(_object) == 0
    return _object is None or object_is_empty


def list_is_not_none_else_none(_object):
    """
    Formats an object as a singleton if not none

    Args:
        object: a python object
    """
    return [_object] if _object is not None else None


def infer_id_from_external_id(kili, asset_id: str, external_id: str, project_id: str):
    """
    Infer asset id from external id

    Args:
        asset_id: asset id
        external_id: external id
        project_id: project id
    """
    if asset_id is None and external_id is None:
        raise Exception(
            'Either provide asset_id or external_id and project_id')
    if asset_id is not None:
        return asset_id
    assets = kili.assets(
        external_id_contains=[external_id], project_id=project_id, fields=['id'], disable_tqdm=True)
    if len(assets) == 0:
        raise Exception(
            f'No asset found with external ID "{external_id}"')
    if len(assets) > 1:
        raise Exception(
            f'Several assets found containing external ID "{external_id}":'
            f' {assets}. Please, use asset ID instead.')
    return assets[0]['id']


def validate_category_search_query(query):
    """Validate the category search query
    Args:
        query: the query to parse

    Raises:
        ValueError: if `query` is invalid
    """
    operator = pp.oneOf(">= <= > < ==")
    number = pp.pyparsing_common.number()
    dot = '.'
    word = pp.Word(pp.alphas, pp.alphanums + "_-*")
    identifier = word + dot + word + dot + 'count'
    condition = identifier + operator + number

    expr = pp.infixNotation(condition, [
        ("AND", 2, pp.opAssoc.LEFT, ),
        ("OR", 2, pp.opAssoc.LEFT, ),
    ])
    try:
        expr.parseString(query, parseAll=True)
    except pp.ParseException as error:
        raise ValueError(f'Invalid category search query: {query}') from error



"""
Fragments of project mutations
"""

PROJECT_FRAGMENT = '''
id
jsonInterface
title
roles {
    user {
      id
      email
    }
    role
}
'''

PROJECT_FRAGMENT_ID = '''
id
'''

PROJECT_FRAGMENT_PUBLIC_TOKEN = '''
id
publicToken
'''

ROLE_FRAGMENT = '''
user {
  id
  email
}
role
'''



"""
Helpers for the project mutations
"""


def verify_argument_ranges(consensus_tot_coverage: int,
                           min_consensus_size: int,
                           review_coverage: int):
    """
    Ensure that all arguments are set in a correct range or raise
    a ValueError otherwise
    """
    if consensus_tot_coverage is not None and (
       consensus_tot_coverage < 0 or consensus_tot_coverage > 100):
        raise ValueError(
            'argument "consensus_tot_coverage" must be comprised between 0 and 100')
    if min_consensus_size is not None and (
       min_consensus_size < 1 or min_consensus_size > 10):
        raise ValueError(
            'argument "min_consensus_size" must be comprised between 1 and 10')
    if review_coverage is not None and (
       review_coverage < 0 or review_coverage > 100):
        raise ValueError(
            'argument "review_coverage" must be comprised between 0 and 100')



"""Project version queries."""

from typing import Generator, List, Optional, Union
import warnings

from typeguard import typechecked


from ...helpers import Compatible, format_result, fragment_builder
from .queries import gql_project_version, GQL_PROJECT_VERSION_COUNT
from ...types import ProjectVersion as ProjectVersionType
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesProjectVersion:
    """Set of ProjectVersion queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v2'])
    @typechecked
    def project_version(
            self,
            first: Optional[int] = 100,
            skip: Optional[int] = 0,
            fields: List[str] = [
                'createdAt',
                'id',
                'content',
                'name',
                'project',
                'projectId'],
            project_id: str = None,
            disable_tqdm: bool = False,
            as_generator: bool = False) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of project versions respecting a set of criteria.

        Args:
            fields: All the fields to request among the possible fields for the project versions
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#projectVersions) for all possible fields.
            first: Number of project versions to query
            project_id: Filter on Id of project
            skip: Number of project versions to skip (they are ordered by their date
                of creation, first to last).
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the project versions is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """

        count_args = {"project_id": project_id}
        disable_tqdm = disable_tqdm or as_generator
        payload_query = {
            'where': {
                'projectId': project_id,
            },
        }
        project_versions_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_project_versions,
            count_args,
            self._query_project_versions,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return project_versions_generator
        return list(project_versions_generator)

    def _query_project_versions(self,
                                skip: int,
                                first: int,
                                payload: dict,
                                fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_project_version = gql_project_version(
            fragment_builder(fields, ProjectVersionType))
        result = self.auth.client.execute(_gql_project_version, payload)
        return format_result('data', result)

    @Compatible(['v2'])
    @typechecked
    def count_project_versions(self, project_id: str) -> int:
        """Count the number of project versions.

        Args:
            project_id: Filter on ID of project

        Returns:
            The number of project versions with the parameters provided
        """
        variables = {
            'where': {'projectId': project_id},
        }
        result = self.auth.client.execute(GQL_PROJECT_VERSION_COUNT, variables)
        count = format_result('data', result)
        return count



"""
Test mutations with pytest
"""
import os
import json
import time
import shutil
import tempfile
from unittest import TestCase
import unittest
import uuid

import pytest
from kili.client import Kili
from kili.mutations.asset.helpers import get_file_mimetype, process_append_many_to_dataset_parameters, process_content
from kili.mutations.asset.queries import GQL_APPEND_MANY_FRAMES_TO_DATASET
import requests

api_key = os.getenv('KILI_USER_API_KEY')
api_endpoint = os.getenv('KILI_API_ENDPOINT')
kili = Kili(api_key=api_key, api_endpoint=api_endpoint)


tester = TestCase()

@pytest.fixture
def create_video_project():
    """
    Create a video project
    """
    with open('./test/mutations/interface.json', 'r') as file:
        interface = json.load(file)
    project = kili.create_project(
        input_type='FRAME', json_interface=interface, title='Test project | video import')
    return project['id']

class LocalDownloader():

    def __init__(self, directory):
        self.directory = directory

    def __call__(self, url):
        content = requests.get(url)
        name = os.path.basename(url)
        path = os.path.join(self.directory, f'{str(uuid.uuid4())}-{name}')
        with open(path, 'wb') as file:
            file.write(content.content)
        return path


TIMEOUT = 10
RETRY = 1
BASE_TEST_CASES = [
    {
        'case': 'uploading a video with a url',
        'content': ['https://storage.googleapis.com/label-public-staging/presales/industry_small.mp4'],
        'expected_processing_parameters': {
            'shouldKeepNativeFrameRate': True,
            'framesPlayedPerSecond': 25,
            'shouldUseNativeVideo': True
        }
    },
    {
        'case': 'uploading a video as a list of image urls',
        'json_content': [["https://storage.googleapis.com/label-public-staging/video1/video1-img000001.jpg",
                            "https://storage.googleapis.com/label-public-staging/video1/video1-img000002.jpg"]],
        'expected_processing_parameters': {
            'shouldKeepNativeFrameRate': False,
            'framesPlayedPerSecond': 30,
            'shouldUseNativeVideo': False
        }
    },
    {
        'case': 'uploading a video into frames from url',
        'content': ['https://storage.googleapis.com/label-public-staging/presales/industry_small.mp4'],
        'expected_processing_parameters': {
            'shouldKeepNativeFrameRate': True,
            'shouldUseNativeVideo': False,
            'framesPlayedPerSecond': 25,
        },
        'json_metadata': [{
            'processingParameters': {
                'shouldKeepNativeFrameRate': True,
                'shouldUseNativeVideo': False,
            }
        }],
    },
    
]
CUSTOM_VALUE_CASES = [
    {
        'case': 'uploading a video with a url',
        'content': ['https://storage.googleapis.com/label-public-staging/presales/industry_small.mp4'],
        'expected_processing_parameters': {
            'shouldKeepNativeFrameRate': False,
            'framesPlayedPerSecond': 40,
            'shouldUseNativeVideo': True
        },
        'json_metadata': [{
            'processingParameters': {
                'shouldKeepNativeFrameRate': False,
                'framesPlayedPerSecond': 40,
            }
        }]
    },
]
TEST_CASES = BASE_TEST_CASES
TEST_CASES.extend([{**x, 'local': True} for x in BASE_TEST_CASES])
TEST_CASES.extend(CUSTOM_VALUE_CASES)


def test_upload_video(create_video_project, tmpdir):
    """
    Test upload of video in different ways
    """
    project_id = create_video_project
    assert project_id != '', 'Project not successfully created'

    downloader = LocalDownloader(tmpdir)

    for i, test_case in enumerate(TEST_CASES):
        external_id = f'video_case_{i+1}'
        content = test_case.get('content', None)
        expected_parameters = test_case.get(
            'expected_processing_parameters', {})
        json_metadata = test_case.get('json_metadata', None)
        json_content = test_case.get('json_content', None)
        local = test_case.get('local', False)
        case = test_case.get('case', 'Unknown case') + f' | {"local" if local else "cloud"}'
        print('case', case)
        if local:
            if content is not None:
                content = list(map(downloader, content))
            else:
                json_content = list(map(lambda jc: [downloader(x) for x in jc], json_content))


        kili.append_many_to_dataset(
            project_id=project_id,
            content_array=content,
            json_metadata_array=json_metadata,
            json_content_array=json_content,
            external_id_array=[external_id]
        )

        tstart = time.time()
        n_assets = kili.count_assets(project_id=project_id)
        while not n_assets == i+1 and time.time() - tstart < TIMEOUT:
            time.sleep(RETRY)
            n_assets = kili.count_assets(project_id=project_id)
        assert n_assets == i+1, 'Asset upload failed'

        asset_uploaded = kili.assets(
            project_id=project_id, external_id_contains=[external_id])
        processed_parameters = asset_uploaded[0]['jsonMetadata']['processingParameters']
        tester.assertDictEqual(expected_parameters, processed_parameters), f'{case}, got {processed_parameters}, expected {expected_parameters}'

class TestMimeType():
    """
    Tests if the mime type is the correct one
    """

    def should_have_right_mimetype(self, content_array, json_content_array, expected_mimetype):
        mimetype = get_file_mimetype(content_array, json_content_array)
        assert mimetype == expected_mimetype, f'Bad mimetype {mimetype}'

    def test_contents_empty(self):
        content_array = None
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, None)

    def test_mimetype_url(self, tmpdir):
        url = 'https://storage.googleapis.com/label-public-staging/car/car_1.jpg'
        content_array = [url]
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, None)

    def test_mimetype_image(self, tmpdir):
        url = 'https://storage.googleapis.com/label-public-staging/car/car_1.jpg'
        downloader = LocalDownloader(tmpdir)
        path = downloader(url)
        content_array = [path]
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, 'image/jpeg')

    def test_mimetype_geotiff(self, tmpdir):
        url = 'https://storage.googleapis.com/label-public-staging/geotiffs/bogota.tif'
        downloader = LocalDownloader(tmpdir)
        path = downloader(url)
        content_array = [path]
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, 'image/tiff')
    
    def test_cannot_upload_mp4_to_image_project(self):
        path = './test.mp4'
        content_array = [path]
        json_content_array = None
        processed_content = process_content('IMAGE', content_array, json_content_array)
        assert(processed_content==[None])

    def test_cannot_upload_png_to_frame_project(self):
        path = './test.png'
        content_array = [path]
        json_content_array = None
        processed_content = process_content('FRAME', content_array, json_content_array)
        assert(processed_content==[None])

    def test_cannot_upload_text_to_pdf_project(self):
        path = 'Hello world'
        content_array = [path]
        json_content_array = None
        processed_content = process_content('PDF', content_array, json_content_array)
        assert(processed_content==[None])

    @pytest.mark.xfail(raises=FileNotFoundError)
    def test_can_upload_png_to_image_project(self):
        path = './test.png'
        content_array = [path]
        json_content_array = None
        process_content('IMAGE', content_array, json_content_array)


class TestUploadTiff(unittest.TestCase):
    """
    Test functions for uploading geotiff
    """

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def test_geotiff_upload_properties(self):
        url = 'https://storage.googleapis.com/label-public-staging/geotiffs/bogota.tif'
        downloader = LocalDownloader(self.test_dir)
        path = downloader(url)
        input_type = 'IMAGE'
        content_array = [path]
        external_id_array = ['bogota']
        is_honeypot_array = None
        status_array = None
        json_content_array = None
        json_metadata_array = None
        properties, upload_type, request = process_append_many_to_dataset_parameters(
            input_type,
            content_array,
            external_id_array,
            is_honeypot_array,
            status_array,
            json_content_array,
            json_metadata_array,
        )
        assert properties['json_metadata_array'] == ['{}']
        assert properties['content_array'][0].startswith(
            'data:image/tiff;base64,SUkqAAgABABre')
        assert properties['external_id_array'] == ['bogota']
        assert upload_type == 'GEO_SATELLITE', 'uploadType do not match'
        assert request == GQL_APPEND_MANY_FRAMES_TO_DATASET, 'Requests do not match'


"""API authentication module"""

import os
import warnings
from datetime import datetime, timedelta

import requests

from . import __version__
from .graphql_client import GraphQLClient
from .helpers import format_result
from .queries.api_key import QueriesApiKey
from .queries.user.queries import GQL_ME

MAX_RETRIES = 20

warnings.filterwarnings("default", module='kili', category=DeprecationWarning)


def get_version_without_patch(version):
    """Return the version of Kili API removing the patch version.

    Args:
        version
    """
    return '.'.join(version.split('.')[:-1])


class KiliAuth:
    """
    from kili.client import Kili
    kili = Kili(api_key=api_key)
    assets = kili.assets(project_id=project_id)
    """

    def __init__(self,
                 api_key=os.getenv('KILI_USER_API_KEY'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/v2/graphql',
                 verify=True):
        self.session = requests.Session()

        self.verify = verify

        if api_endpoint and 'v1/graphql' in api_endpoint:
            # pylint: disable=line-too-long
            message = 'We are migrating the API to enhance our service,' \
                ' please use the new endpoint https://cloud.kili-technology.com/api/label/v2/graphql' \
                ' (or None), the former endpoint call will be deprecated on February 15th 2021'
            warnings.warn(message, DeprecationWarning)
        try:
            self.check_versions_match(api_endpoint)
        except:  # pylint: disable=bare-except
            message = 'We could not check the version, there might be a version' \
                'mismatch or the app might be in deployment'
            warnings.warn(message, UserWarning)

        adapter = requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES)
        self.session.mount('https://', adapter)
        self.session.mount('http://', adapter)
        self.client = GraphQLClient(
            api_endpoint, self.session, verify=self.verify)
        if api_key is None:
            message = 'You need to provide an API KEY to connect.' \
                ' Visit https://docs.kili-technology.com/reference/creating-an-api-key'
            warnings.warn(message, UserWarning)
        self.client.inject_token('X-API-Key: ' + api_key)

        user = self.get_user()

        if (user is None or user['id'] is None or user['email'] is None):
            raise Exception('No user attached to the API key was found')

        self.user_id = user['id']
        self.user_email = user['email']
        self.check_expiry_of_key_is_close(api_key)

    def __del__(self):
        self.session.close()

    def check_versions_match(self, api_endpoint):
        """Check that the versions of Kili Python SDK and Kili API are the same

        Args:
            api_endpoint: url of the Kili API
        """
        url = api_endpoint.replace('/graphql', '/version')
        response = requests.get(url, verify=self.verify).json()
        version = response['version']
        if get_version_without_patch(version) != get_version_without_patch(__version__):
            message = 'Kili Python SDK version should match with Kili API version.\n' + \
                      f'Please install version: "pip install kili=={version}"'
            warnings.warn(message, UserWarning)

    def check_expiry_of_key_is_close(self, api_key):
        """Check that the expiration date of the api_key is not too close.

        Args:
            api_key: key used to connect to the Kili API
        """
        duration_days = 365
        warn_days = 30
        queries = QueriesApiKey(self)
        key_object = queries.api_keys(api_key=api_key, fields=[
                                      'createdAt'], disable_tqdm=True)
        key_creation = datetime.strptime(
            key_object[0]['createdAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
        key_expiry = key_creation + timedelta(days=duration_days)
        key_remaining_time = key_expiry - datetime.now()
        key_soon_deprecated = key_remaining_time < timedelta(days=warn_days)
        if key_soon_deprecated:
            message = f"""
Your api key will be deprecated on {key_expiry:%Y-%m-%d}.
You should generate a new one on My account > API KEY."""
            warnings.warn(message, UserWarning)

    def get_user(self):
        """Get the current user from the api_key provided"""
        result = self.client.execute(GQL_ME)
        return format_result('data', result)




"""Asset queries."""

from typing import Generator, List, Optional, Union
import warnings

from typeguard import typechecked
import pandas as pd

from ...helpers import (Compatible, deprecate, format_result,
                        fragment_builder, validate_category_search_query)
from .queries import gql_assets, GQL_ASSETS_COUNT
from ...types import Asset as AssetType
from ...orm import Asset
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesAsset:
    """
    Set of Asset queries
    """
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v1', 'v2'])
    @typechecked
    @deprecate(removed_in="2.115")
    def assets(self,
               asset_id: Optional[str] = None,
               project_id: Optional[str] = None,
               skip: int = 0,
               fields: List[str] = ['content',
                                    'createdAt',
                                    'externalId',
                                    'id',
                                    'isHoneypot',
                                    'jsonMetadata',
                                    'labels.author.id',
                                    'labels.author.email',
                                    'labels.createdAt',
                                    'labels.id',
                                    'labels.jsonResponse',
                                    'skipped',
                                    'status'],
               asset_id_in: Optional[List[str]] = None,
               consensus_mark_gt: Optional[float] = None,
               consensus_mark_lt: Optional[float] = None,
               disable_tqdm: bool = False,
               external_id_contains: Optional[List[str]] = None,
               first: Optional[int] = None,
               format: Optional[str] = None,  # pylint: disable=redefined-builtin
               honeypot_mark_gt: Optional[float] = None,
               honeypot_mark_lt: Optional[float] = None,
               label_author_in: Optional[List[str]] = None,
               label_consensus_mark_gt: Optional[float] = None,
               label_consensus_mark_lt: Optional[float] = None,
               label_created_at: Optional[str] = None,
               label_created_at_gt: Optional[str] = None,
               label_created_at_lt: Optional[str] = None,
               label_honeypot_mark_gt: Optional[float] = None,
               label_honeypot_mark_lt: Optional[float] = None,
               label_json_response_contains: Optional[List[str]] = None,
               label_type_in: Optional[List[str]] = None,
               metadata_where: Optional[dict] = None,
               skipped: Optional[bool] = None,
               status_in: Optional[List[str]] = None,
               updated_at_gte: Optional[str] = None,
               updated_at_lte: Optional[str] = None,
               as_generator: bool = False,
               label_category_search: Optional[str] = None,
               ) -> Union[List[dict], Generator[dict, None, None], pd.DataFrame]:
        # pylint: disable=line-too-long
        """Get an asset list, an asset generator or a pandas DataFrame that match a set of constraints.

        Args:
            asset_id: Identifier of the asset to retrieve.
            asset_id_in: A list of the IDs of the assets to retrieve.
            project_id: Identifier of the project.
            skip: Number of assets to skip (they are ordered by their date of creation, first to last).
            fields: All the fields to request among the possible fields for the assets.
                    See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.
            first: Maximum number of assets to return.
            consensus_mark_gt: Minimum amount of consensus for the asset.
            consensus_mark_lt: Maximum amount of consensus for the asset.
            external_id_contains: Returned assets have an external id that belongs to that list, if given.
            metadata_where: Filters by the values of the metadata of the asset.
            honeypot_mark_gt: Minimum amount of honeypot for the asset.
            honeypot_mark_lt : Maximum amount of honeypot for the asset.
            status_in: Returned assets should have a status that belongs to that list, if given.
                Possible choices: `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            label_type_in: Returned assets should have a label whose type belongs to that list, if given.
            label_author_in: Returned assets should have a label whose status belongs to that list, if given.
            label_consensus_mark_gt: Returned assets should have a label whose consensus is greater than this number.
            label_consensus_mark_lt: Returned assets should have a label whose consensus is lower than this number.
            label_created_at: Returned assets should have a label whose creation date is equal to this date.
            label_created_at_gt: Returned assets should have a label whose creation date is greater than this date.
            label_created_at_lt: Returned assets should have a label whose creation date is lower than this date.
            label_json_response_contains: Returned assets should have a substring of the label's jsonResponse
                    that belongs to that list, if given.
            label_honeypot_mark_gt: Returned assets should have a label whose honeypot is greater than this number
            label_honeypot_mark_lt: Returned assets should have a label whose honeypot is lower than this number
            skipped: Returned assets should be skipped
            updated_at_gte: Returned assets should have a label whose update date is greated or equal to this date.
            updated_at_lte: Returned assets should have a label whose update date is lower or equal to this date.
            format: If equal to 'pandas', returns a pandas DataFrame
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the assets is returned.
            label_category_search: Returned assets should have a label that follows this category search query.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Example:
            ```
            # returns the assets list of the project
            >>> kili.assets(project_id=project_id)
            >>> kili.assets(asset_id=asset_id)
            # returns a generator of the project assets
            >>> kili.assets(project_id=project_id, as_generator=True)
            ```

        !!! example "How to filter based on Metadata"
            - `metadata_where = {key1: "value1"}` to filter on assets whose metadata
                have key "key1" with value "value1"
            - `metadata_where = {key1: ["value1", "value2"]}` to filter on assets whose metadata
                have key "key1" with value "value1" or value "value2
            - `metadata_where = {key2: [2, 10]}` to filter on assets whose metadata
                have key "key2" with a value between 2 and 10.

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators

            Example:

                label_category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                label_category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                label_category_search = `(JOB_CLASSIF.CATEGORY_A.count == 1 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
        """
        if format == "pandas" and as_generator:
            raise ValueError(
                "Argument values as_generator==True and format==\"pandas\" are not compatible.")

        saved_args = locals()
        count_args = {k: v for (k, v) in saved_args.items()
                      if k not in ['skip', 'first', 'disable_tqdm', 'format', 'fields', 'self', 'as_generator']}

        # using tqdm with a generator is messy, so it is always disabled
        disable_tqdm = disable_tqdm or as_generator
        if label_category_search:
            validate_category_search_query(label_category_search)

        payload_query = {
            'where': {
                'id': asset_id,
                'project': {
                    'id': project_id,
                },
                'externalIdIn': external_id_contains,
                'statusIn': status_in,
                'consensusMarkGte': consensus_mark_gt,
                'consensusMarkLte': consensus_mark_lt,
                'honeypotMarkGte': honeypot_mark_gt,
                'honeypotMarkLte': honeypot_mark_lt,
                'idIn': asset_id_in,
                'metadata': metadata_where,
                'label': {
                    'typeIn': label_type_in,
                    'authorIn': label_author_in,
                    'consensusMarkGte': label_consensus_mark_gt,
                    'consensusMarkLte': label_consensus_mark_lt,
                    'createdAt': label_created_at,
                    'createdAtGte': label_created_at_gt,
                    'createdAtLte': label_created_at_lt,
                    'honeypotMarkGte': label_honeypot_mark_gt,
                    'honeypotMarkLte': label_honeypot_mark_lt,
                    'jsonResponseContains': label_json_response_contains,
                    'search': label_category_search
                },
                'skipped': skipped,
                'updatedAtGte': updated_at_gte,
                'updatedAtLte': updated_at_lte,
            },
        }

        asset_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_assets,
            count_args,
            self._query_assets,
            payload_query,
            fields,
            disable_tqdm
        )

        if format == "pandas":
            return pd.DataFrame(list(asset_generator))
        if as_generator:
            return asset_generator
        return list(asset_generator)

    def _query_assets(self,
                      skip: int,
                      first: int,
                      payload: dict,
                      fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_assets = gql_assets(fragment_builder(fields, AssetType))
        result = self.auth.client.execute(_gql_assets, payload)
        assets = format_result('data', result, Asset)
        return assets

    @Compatible(['v1', 'v2'])
    @typechecked
    @deprecate(removed_in="2.115")
    def count_assets(self, asset_id: Optional[str] = None,
                     project_id: Optional[str] = None,
                     asset_id_in: Optional[List[str]] = None,
                     external_id_contains: Optional[List[str]] = None,
                     metadata_where: Optional[dict] = None,
                     status_in: Optional[List[str]] = None,
                     consensus_mark_gt: Optional[float] = None,
                     consensus_mark_lt: Optional[float] = None,
                     honeypot_mark_gt: Optional[float] = None,
                     honeypot_mark_lt: Optional[float] = None,
                     label_type_in: Optional[List[str]] = None,
                     label_author_in: Optional[List[str]] = None,
                     label_consensus_mark_gt: Optional[float] = None,
                     label_consensus_mark_lt: Optional[float] = None,
                     label_created_at: Optional[str] = None,
                     label_created_at_gt: Optional[str] = None,
                     label_created_at_lt: Optional[str] = None,
                     label_honeypot_mark_gt: Optional[float] = None,
                     label_honeypot_mark_lt: Optional[float] = None,
                     label_json_response_contains: Optional[List[str]] = None,
                     skipped: Optional[bool] = None,
                     updated_at_gte: Optional[str] = None,
                     updated_at_lte: Optional[str] = None,
                     label_category_search: Optional[str] = None) -> int:
        """Count and return the number of assets with the given constraints.

        Parameters beginning with 'label_' apply to labels, others apply to assets.

        Args:
            asset_id: The unique id of the asset to retrieve.
            asset_id_in: A list of the ids of the assets to retrieve.
            project_id: Identifier of the project
            external_id_contains: Returned assets should have an external id
                that belongs to that list, if given.
            metadata_where: Filters by the values of the metadata of the asset.
            status_in: Returned assets should have a status that belongs to that list, if given.
                Possible choices: `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            consensus_mark_gt: Minimum amount of consensus for the asset.
            consensus_mark_lt: Maximum amount of consensus for the asset.
            honeypot_mark_gt: Minimum amount of honeypot for the asset.
            honeypot_mark_lt: Maximum amount of consensus for the asset.
            label_type_in: Returned assets should have a label
                whose type belongs to that list, if given.
            label_author_in: Returned assets should have a label
                whose status belongs to that list, if given.
            label_consensus_mark_gt: Returned assets should have a label
                whose consensus is greater than this number.
            label_consensus_mark_lt: Returned assets should have a label
                whose consensus is lower than this number.
            label_created_at: Returned assets should have a label
                whose creation date is equal to this date.
            label_created_at_gt: Returned assets should have a label
                whose creation date is greater than this date.
            label_created_at_lt: Returned assets should have a label
                whose creation date is lower than this date.
            label_honeypot_mark_gt: Returned assets should have a label
                whose honeypot is greater than this number.
            label_honeypot_mark_lt: Returned assets should have a label
                whose honeypot is lower than this number.
            label_json_response_contains: Returned assets should have
                a substring of the label's jsonResponse that belongs
                to that list, if given.
            skipped: Returned assets should be skipped
            updated_at_gte: Returned assets should have a label
                whose update date is greated or equal to this date.
            updated_at_lte: Returned assets should have a label
                whose update date is lower or equal to this date.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.count_assets(project_id=project_id)
            250
            >>> kili.count_assets(asset_id=asset_id)
            1

        !!! example "How to filter based on Metadata"
            - `metadata_where = {key1: "value1"}` to filter on assets whose metadata
                have key "key1" with value "value1"
            - `metadata_where = {key1: ["value1", "value2"]}` to filter on assets whose metadata
                have key "key1" with value "value1" or value "value2
            - `metadata_where = {key2: [2, 10]}` to filter on assets whose metadata
                have key "key2" with a value between 2 and 10.
        """

        if label_json_response_contains is not None:
            message = """
                The field `label_json_response_contains` is deprecated since: 2.113
                It will be removed in: 2.115
                Please use `label_category_search` to filter based on categories in labels
                """
            warnings.warn(message, DeprecationWarning)

        if label_category_search:
            validate_category_search_query(label_category_search)

        variables = {
            'where': {
                'id': asset_id,
                'project': {
                    'id': project_id,
                },
                'externalIdIn': external_id_contains,
                'statusIn': status_in,
                'consensusMarkGte': consensus_mark_gt,
                'consensusMarkLte': consensus_mark_lt,
                'honeypotMarkGte': honeypot_mark_gt,
                'honeypotMarkLte': honeypot_mark_lt,
                'idIn': asset_id_in,
                'metadata': metadata_where,
                'label': {
                    'typeIn': label_type_in,
                    'authorIn': label_author_in,
                    'consensusMarkGte': label_consensus_mark_gt,
                    'consensusMarkLte': label_consensus_mark_lt,
                    'createdAt': label_created_at,
                    'createdAtGte': label_created_at_gt,
                    'createdAtLte': label_created_at_lt,
                    'honeypotMarkGte': label_honeypot_mark_gt,
                    'honeypotMarkLte': label_honeypot_mark_lt,
                    'jsonResponseContains': label_json_response_contains,
                    'search': label_category_search
                },
                'skipped': skipped,
                'updatedAtGte': updated_at_gte,
                'updatedAtLte': updated_at_lte,
            }
        }
        result = self.auth.client.execute(GQL_ASSETS_COUNT, variables)
        count = format_result('data', result)
        return count



"""Organization mutations."""

import json
from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from .queries import (GQL_CREATE_ORGANIZATION,
                      GQL_UPDATE_PROPERTIES_IN_ORGANIZATION)


class MutationsOrganization:
    """Set of Organization mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initializes the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(['v1', 'v2'])
    @typechecked
    def create_organization(self, name: str, address: str,
                            zip_code: str, city: str, country: str):
        """Create an organization.

        Each user must be linked to an organization

        Args:
            name :
            address :
            zip_code :
            city :
            country :

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            'data': {'name': name,
                     'address': address,
                     'zipCode': zip_code,
                     'city': city,
                     'country': country}
        }
        result = self.auth.client.execute(GQL_CREATE_ORGANIZATION, variables)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def update_properties_in_organization(self,
                                          organization_id: str,
                                          name: Optional[str] = None,
                                          address: Optional[str] = None,
                                          zip_code: Optional[str] = None,
                                          city: Optional[str] = None,
                                          country: Optional[str] = None,
                                          license: Optional[dict] = None):  # pylint: disable=redefined-builtin
        """Modify an organization.

        Args:
            organization_id :
            name :
            address :
            license :
            zip_code :
            city :
            country :

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        license_str = None if not license else json.dumps(license)
        variables = {
            'id': organization_id
        }
        if name is not None:
            variables['name'] = name
        if address is not None:
            variables['address'] = address
        if license_str is not None:
            variables['license'] = license_str
        if zip_code is not None:
            variables['zipCode'] = zip_code
        if city is not None:
            variables['city'] = city
        if country is not None:
            variables['country'] = country
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_ORGANIZATION, variables)
        return format_result('data', result)



"""Tests for category search validation"""

import pytest

from kili.helpers import validate_category_search_query

TEST_CASES = [
    {
        "case": "user does not provide job",
        "query": "class_A.count > 0",
        "raise_error": True,
    },
    {
        "case": "user does not write count",
        "query": "JOB.class_A > 0",
        "raise_error": True,
    },
    {
        "case": "user can provide expressions separated by OR and AND",
        "query": "JOB.class_A.count > 0 AND JOB2.class_B.count == 3 OR JOB2.class_C.count < 3",
        "raise_error": False,
    },
    {
        "case": "user can provide expressions with parenthesis",
        "query": "JOB.class_A.count > 0 AND (JOB2.class_B.count == 3 OR JOB2.class_C.count < 3)",
        "raise_error": False,
    },
    {
        "case": "user can add spaces where it does not break the query",
        "query": "JOB. class_A.  count > 0 AND      JOB2.class_B. count ==   3",
        "raise_error": False,
    },
    {
        "case": "user did not close a parenthesis",
        "query": "JOB.class_A.count > 0 AND (JOB2.class_B.count == 3 OR JOB2.class_C.count < 3",
        "raise_error": True,
    },
    {
        "case": "user can have complex job and ategory name",
        "query": "JoB46_Hhzef*bzf66.class_Auzf657bdh----_.count > 0 ",
        "raise_error": False,
    }]


def test_category_search_queries():
    for test in TEST_CASES:
        case, query, raise_error = test.get('case'), test.get(
            'query'), test.get('raise_error'),
        if raise_error:
            with pytest.raises(ValueError):
                validate_category_search_query(query)
        else:
            validate_category_search_query(query)



"""Deprecation tests"""

import inspect
from kili import client
from kili import __version__


def test_catch_deprecated_elements_to_remove():
    """Raise an error if a function has a deprecation tag of a version
    that is bigger that the version that is released."""
    functions = inspect.getmembers(client.Kili, inspect.isfunction)
    functions_warning = []
    for function_name, function in functions:
        if hasattr(function, 'removed_in'):
            deprecation_version = getattr(function, 'removed_in')
            if __version__.split('.')[:2] >= deprecation_version.split('.'):
                functions_warning.append(function_name)
    assert (len(functions_warning) ==
            0), f'Functions: {[function_name for function_name in functions_warning]}' \
                'are deprecated or have a deprecated argument that should be removed'\
                f'in version {__version__}'



"""User queries."""

from typing import Generator, List, Optional, Union
import warnings

from typeguard import typechecked


from ...helpers import Compatible, format_result, fragment_builder
from .queries import gql_users, GQL_USERS_COUNT
from ...types import User
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesUser:
    """Set of User queries."""
    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v1', 'v2'])
    @typechecked
    def users(self,
              api_key: Optional[str] = None,
              email: Optional[str] = None,
              organization_id: Optional[str] = None,
              fields: List[str] = ['email', 'id', 'firstname', 'lastname'],
              first: int = 100,
              skip: int = 0,
              disable_tqdm: bool = False,
              as_generator: bool = False) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of users given a set of criteria

        Args:
            api_key: Query an user by its API Key
            email: Email of the user
            organization_id: Identifier of the user's organization
            fields: All the fields to request among the possible fields for the users.
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#user) for all possible fields.
            first: Maximum number of users to return
            skip: Number of skipped users (they are ordered by creation date)
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the users is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            ```
            # List all users in my organization
            >>> organization = kili.organizations()
            >>> organization_id = organizations[0]['id]
            >>> kili.users(organization_id=organization_id)
            ```
        """

        count_args = {"organization_id": organization_id}
        disable_tqdm = disable_tqdm or as_generator or (
            api_key or email) is not None
        payload_query = {
            'where': {
                'apiKey': api_key,
                'email': email,
                'organization': {
                    'id': organization_id,
                }
            }
        }

        users_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_users,
            count_args,
            self._query_users,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return users_generator
        return list(users_generator)

    def _query_users(self,
                     skip: int,
                     first: int,
                     payload: dict,
                     fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_users = gql_users(fragment_builder(fields, User))
        result = self.auth.client.execute(_gql_users, payload)
        return format_result('data', result)

    @Compatible(['v1', 'v2'])
    @typechecked
    def count_users(self,
                    organization_id: Optional[str] = None) -> int:
        """Get user count based on a set of constraints.

        Args:
            organization_id: Identifier of the user's organization

        Returns:
            The number of organizations with the parameters provided
        """
        variables = {
            'where': {
                'organization': {
                    'id': organization_id,
                }
            }
        }
        result = self.auth.client.execute(GQL_USERS_COUNT, variables)
        return format_result('data', result)



"""
Utils
"""
from typing import Dict, List, Callable, Optional
import time
from tqdm import tqdm

from kili.constants import MUTATION_BATCH_SIZE, THROTTLING_DELAY
from kili.helpers import GraphQLError

# pylint: disable=too-many-arguments,too-many-locals


def row_generator_from_paginated_calls(
    skip: int,
    first: int,
    count_method: Callable[..., int],
    count_kwargs: dict,
    paged_call_method: Callable[..., List[dict]],
    paged_call_payload: dict,
    fields: List[str],
    disable_tqdm: bool,
):
    """
    Builds a row generator from paginated calls.

    Args:
        skip: Number of assets to skip (they are ordered by their date of creation, first to last).
        first: Maximum number of assets to return.
        count_method: Callable returning the number of available assets given `count_args`.
        count_kwargs: Keyword arguments passed to the `count_method`.
        paged_call_method: Callable returning the list of samples.
        paged_call_payload: Payload for the GraphQL query.
        fields: The list of strings to retrieved.
        disable_tqdm: If `True`, disables tqdm.
    """
    count_rows_retrieved = 0
    if not disable_tqdm:
        count_rows_available = count_method(**count_kwargs)
        count_rows_queried_total = min(count_rows_available,
                                       first) if first is not None else count_rows_available
    else:
        # dummy value that won't have any impact since tqdm is disabled
        count_rows_queried_total = 1 if first != 0 else 0
    count_rows_query_default = min(100, first or 100)
    throttling_delay = 60 / 250

    if count_rows_queried_total == 0:
        yield from ()
    else:
        with tqdm(total=count_rows_queried_total, disable=disable_tqdm) as pbar:
            while True:
                query_start = time.time()
                rows = paged_call_method(
                    count_rows_retrieved + skip,
                    count_rows_query_default,
                    paged_call_payload,
                    fields,
                )
                query_time = time.time() - query_start

                if query_time < throttling_delay:
                    time.sleep(throttling_delay - query_time)

                if rows is None or len(rows) == 0:
                    break

                for row in rows:
                    yield row

                count_rows_retrieved += len(rows)
                pbar.update(len(rows))
                if first is not None and count_rows_retrieved >= first:
                    break


def batch_iterator_builder(iterable: List, batch_size=MUTATION_BATCH_SIZE):
    """Generate an paginated iterator from a list
    Args:
        iterable: a list to paginate.
        batch_size: the size of the batches to produce
    """
    iterable_length = len(iterable)
    for ndx in range(0, iterable_length, batch_size):
        yield iterable[ndx:min(ndx + batch_size, iterable_length)]


def batch_object_builder(
        properties_to_batch: Dict[str, Optional[list]],
        batch_size: int = MUTATION_BATCH_SIZE) -> Dict[str, Optional[list]]:
    """Generate a paginated iterator for several variables
    Args:
        properties_to_batch: a dictionnary of properties to be batched.
        batch_size: the size of the batches to produce
    """
    if len(list(filter(None, properties_to_batch.values()))) == 0:
        print(properties_to_batch)
        yield properties_to_batch
        return
    number_of_objects = len([v for v in properties_to_batch.values(
    ) if v is not None][0])
    number_of_batches = len(range(0, number_of_objects, batch_size))
    batched_properties = {k: (batch_iterator_builder(v, batch_size) if v is not None
                              else (item for item in [v]*number_of_batches))
                          for k, v in properties_to_batch.items()}
    batch_object_iterator = [dict(zip(batched_properties, t))
                             for t in zip(*batched_properties.values())]
    for batch in batch_object_iterator:
        yield batch


def _mutate_from_paginated_call(self,
                                properties_to_batch: Dict[str, Optional[list]],
                                generate_variables: Callable,
                                request: str,
                                batch_size: int = MUTATION_BATCH_SIZE):
    """Run a mutation by making paginated calls
    Args:
        properties_to_batch: a dictionnary of properties to be batched.
            constants across batch are defined in the generate_variables function
        generate_variables: function that takes batched properties and return
            a graphQL payload for request for this batch
        request: the GraphQL request to call,
        batch_size: the size of the batches to produce
    Example:
        '''
        properties_to_batch={prop1: [0,1], prop2: ['a', 'b']}
        def generate_variables(batched_properties):
            return {
                graphQL_prop1: batched_properties['prop1']
                graphQL_prop2: batched_properties['prop2']
            }
        _mutate_from_paginated_call(
                properties_to_batch=properties_to_batch,
                generate_variables=generate_variables
                request= APPEND_MANY_TO_DATASET
                )
        '''
    """
    results = []
    for batch_number, batch in enumerate(batch_object_builder(properties_to_batch, batch_size)):
        mutation_start = time.time()
        variables = generate_variables(batch)
        result = self.auth.client.execute(request, variables)
        mutation_time = time.time() - mutation_start
        results.append(result)
        if 'errors' in result:
            raise GraphQLError('data', result['errors'], batch_number)
        if mutation_time < THROTTLING_DELAY:
            time.sleep(THROTTLING_DELAY - mutation_time)
    return results



"""Label queries."""

from typing import Generator, List, Optional, Union
import warnings

from typeguard import typechecked
import pandas as pd


from ...helpers import (Compatible, deprecate, format_result,
                        fragment_builder, validate_category_search_query)
from ..asset import QueriesAsset
from ..project import QueriesProject
from .queries import gql_labels, GQL_LABELS_COUNT
from ...constants import NO_ACCESS_RIGHT
from ...types import Label as LabelType
from ...orm import Label
from ...utils.pagination import row_generator_from_paginated_calls


class QueriesLabel:
    """Set of Label queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(['v1', 'v2'])
    @typechecked
    @deprecate(removed_in="2.115")
    def labels(self,
               asset_id: Optional[str] = None,
               asset_status_in: Optional[List[str]] = None,
               asset_external_id_in: Optional[List[str]] = None,
               author_in: Optional[List[str]] = None,
               created_at: Optional[str] = None,
               created_at_gte: Optional[str] = None,
               created_at_lte: Optional[str] = None,
               fields: List[str] = ['author.email', 'author.id', 'id',
                                    'jsonResponse', 'labelType', 'secondsToLabel', 'skipped'],
               first: Optional[int] = None,
               honeypot_mark_gte: Optional[float] = None,
               honeypot_mark_lte: Optional[float] = None,
               id_contains: Optional[List[str]] = None,
               json_response_contains: Optional[List[str]] = None,
               label_id: Optional[str] = None,
               project_id: Optional[str] = None,
               skip: int = 0,
               skipped: Optional[bool] = None,
               type_in: Optional[List[str]] = None,
               user_id: Optional[str] = None,
               disable_tqdm: bool = False,
               as_generator: bool = False,
               category_search: Optional[str] = None,
               ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a label list or a label generator from a project based on a set of criteria.

        Args:
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            json_response_contains: Returned labels should have a substring of the jsonResponse that belongs
                to that list, if given.
            label_id: Identifier of the label.
            project_id: Identifier of the project.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            skipped: Returned labels should have a label which is skipped
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful, else an error message.

        Examples:
            >>> kili.labels(project_id=project_id, fields=['jsonResponse', 'labelOf.externalId']) # returns a list of all labels of a project and their assets external ID
            >>> kili.labels(project_id=project_id, fields=['jsonResponse'], as_generator=True) # returns a generator of all labels of a project

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators

            Example:

                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                category_search = `(JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
        """

        saved_args = locals()
        count_args = {
            k: v
            for (k, v) in saved_args.items()
            if k
            not in [
                'as_generator',
                'disable_tqdm',
                'fields',
                'first',
                'id_contains',
                'self',
                'skip',
            ]
        }

        # using tqdm with a generator is messy, so it is always disabled
        disable_tqdm = disable_tqdm or as_generator

        if category_search:
            validate_category_search_query(category_search)

        payload_query = {
            'where': {
                'id': label_id,
                'asset': {
                    'id': asset_id,
                    'externalIdIn': asset_external_id_in,
                    'statusIn': asset_status_in,
                },
                'project': {
                    'id': project_id,
                },
                'user': {
                    'id': user_id,
                },
                'createdAt': created_at,
                'createdAtGte': created_at_gte,
                'createdAtLte': created_at_lte,
                'authorIn': author_in,
                'honeypotMarkGte': honeypot_mark_gte,
                'honeypotMarkLte': honeypot_mark_lte,
                'idIn': id_contains,
                'jsonResponseContains': json_response_contains,
                'search': category_search,
                'skipped': skipped,
                'typeIn': type_in,
            },
        }

        labels_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_labels,
            count_args,
            self._query_labels,
            payload_query,
            fields,
            disable_tqdm
        )

        if as_generator:
            return labels_generator
        return list(labels_generator)

    def _query_labels(self,
                      skip: int,
                      first: int,
                      payload: dict,
                      fields: List[str]):

        payload.update({'skip': skip, 'first': first})
        _gql_labels = gql_labels(fragment_builder(fields, LabelType))
        result = self.auth.client.execute(_gql_labels, payload)
        return format_result('data', result, Label)

    # pylint: disable=dangerous-default-value
    @typechecked
    def export_labels_as_df(self,
                            project_id: str,
                            fields: List[str] = [
                                'author.email',
                                'author.id',
                                'createdAt',
                                'id',
                                'labelType',
                                'skipped'
                            ],
                            asset_fields: List[str] = [
                                'externalId'
                            ]) -> pd.DataFrame:
        # pylint: disable=line-too-long
        """Get the labels of a project as a pandas DataFrame.

        Args:
            project_id: Identifier of the project
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            asset_fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.

        Returns:
            pandas DataFrame containing the labels.
        """
        projects = QueriesProject(self.auth).projects(project_id)
        assert len(projects) == 1, NO_ACCESS_RIGHT
        assets = QueriesAsset(self.auth).assets(
            project_id=project_id,
            fields=asset_fields + ['labels.' + field for field in fields])
        labels = [dict(label, **dict((f'asset_{key}', asset[key]) for key in asset if key != 'labels'))
                  for asset in assets for label in asset['labels']]
        labels_df = pd.DataFrame(labels)
        return labels_df

    @Compatible(['v1', 'v2'])
    @typechecked
    @deprecate(removed_in="2.115")
    def count_labels(self,
                     asset_id: Optional[str] = None,
                     asset_status_in: Optional[List[str]] = None,
                     asset_external_id_in: Optional[List[str]] = None,
                     author_in: Optional[List[str]] = None,
                     created_at: Optional[str] = None,
                     created_at_gte: Optional[str] = None,
                     created_at_lte: Optional[str] = None,
                     honeypot_mark_gte: Optional[float] = None,
                     honeypot_mark_lte: Optional[float] = None,
                     json_response_contains: Optional[List[str]] = None,
                     label_id: Optional[str] = None,
                     project_id: Optional[str] = None,
                     skipped: Optional[bool] = None,
                     type_in: Optional[List[str]] = None,
                     user_id: Optional[str] = None,
                     category_search: Optional[str] = None) -> int:
        # pylint: disable=line-too-long
        """Get the number of labels for the given parameters.

        Args:
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            json_response_contains: Returned labels should have a substring of the jsonResponse that
                belongs to that list, if given.
            label_id: Identifier of the label.
            project_id: Identifier of the project.
            skipped: Returned labels should have a label which is skipped
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of labels with the parameters provided
        """

        if json_response_contains is not None:
            message = """
                The field `json_response_contains` is deprecated since: 2.113
                It will be removed in: 2.115
                Please use `category_search` to filter based on categories in labels
                """
            warnings.warn(message, DeprecationWarning)

        if category_search:
            validate_category_search_query(category_search)

        variables = {
            'where': {
                'id': label_id,
                'asset': {
                    'id': asset_id,
                    'externalIdIn': asset_external_id_in,
                    'statusIn': asset_status_in,
                },
                'project': {
                    'id': project_id,
                },
                'user': {
                    'id': user_id,
                },
                'createdAt': created_at,
                'createdAtGte': created_at_gte,
                'createdAtLte': created_at_lte,
                'authorIn': author_in,
                'honeypotMarkGte': honeypot_mark_gte,
                'honeypotMarkLte': honeypot_mark_lte,
                'jsonResponseContains': json_response_contains,
                'search': category_search,
                'skipped': skipped,
                'typeIn': type_in,
            }
        }
        result = self.auth.client.execute(GQL_LABELS_COUNT, variables)
        count = format_result('data', result)
        return count
